import numpy as np
from numpy.testing import assert_almost_equal

from yt import load
from yt.frontends.swift.api import SwiftDataset
from yt.testing import ParticleSelectionComparison, requires_file, requires_module
from yt.utilities.on_demand_imports import _h5py as h5py

keplerian_ring = "KeplerianRing/keplerian_ring_0020.hdf5"
EAGLE_6 = "EAGLE_6/eagle_0005.hdf5"


# Combined the tests for loading a file and ensuring the units have been
# implemented correctly to save time on re-loading a dataset
@requires_module("h5py")
@requires_file(keplerian_ring)
def test_non_cosmo_dataset():
    ds = load(keplerian_ring)
    assert type(ds) is SwiftDataset

    field = ("gas", "density")
    ad = ds.all_data()
    yt_density = ad[field]
    yt_coords = ad[field[0], "position"]

    # load some data the old fashioned way
    fh = h5py.File(ds.parameter_filename, mode="r")
    part_data = fh["PartType0"]

    # set up a conversion factor by loading the unit mas and unit length in cm,
    # and then converting to proper coordinates
    units = fh["Units"]
    units = dict(units.attrs)
    density_factor = float(units["Unit mass in cgs (U_M)"])
    density_factor /= float(units["Unit length in cgs (U_L)"]) ** 3

    # now load the raw density and coordinates
    raw_density = part_data["Density"][:].astype("float64") * density_factor
    raw_coords = part_data["Coordinates"][:].astype("float64")
    fh.close()

    # sort by the positions - yt often loads in a different order
    ind_raw = np.lexsort((raw_coords[:, 2], raw_coords[:, 1], raw_coords[:, 0]))
    ind_yt = np.lexsort((yt_coords[:, 2], yt_coords[:, 1], yt_coords[:, 0]))
    raw_density = raw_density[ind_raw]
    yt_density = yt_density[ind_yt]

    # make sure we are comparing fair units
    assert str(yt_density.units) == "g/cm**3"

    # make sure the actual values are the same
    assert_almost_equal(yt_density.d, raw_density)


@requires_module("h5py")
@requires_file(keplerian_ring)
def test_non_cosmo_dataset_selection():
    ds = load(keplerian_ring)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()


@requires_module("h5py")
@requires_file(EAGLE_6)
def test_cosmo_dataset():
    ds = load(EAGLE_6)
    assert type(ds) is SwiftDataset

    field = ("gas", "density")
    ad = ds.all_data()
    yt_density = ad[field]
    yt_coords = ad[field[0], "position"]

    # load some data the old fashioned way
    fh = h5py.File(ds.parameter_filename, mode="r")
    part_data = fh["PartType0"]

    # set up a conversion factor by loading the unit mas and unit length in cm,
    # and then converting to proper coordinates
    units = fh["Units"]
    units = dict(units.attrs)
    density_factor = float(units["Unit mass in cgs (U_M)"])
    density_factor /= float(units["Unit length in cgs (U_L)"]) ** 3

    # add the redshift factor
    header = fh["Header"]
    header = dict(header.attrs)
    density_factor *= (1.0 + float(header["Redshift"])) ** 3

    # now load the raw density and coordinates
    raw_density = part_data["Density"][:].astype("float64") * density_factor
    raw_coords = part_data["Coordinates"][:].astype("float64")
    fh.close()

    # sort by the positions - yt often loads in a different order
    ind_raw = np.lexsort((raw_coords[:, 2], raw_coords[:, 1], raw_coords[:, 0]))
    ind_yt = np.lexsort((yt_coords[:, 2], yt_coords[:, 1], yt_coords[:, 0]))
    raw_density = raw_density[ind_raw]
    yt_density = yt_density[ind_yt]

    # make sure we are comparing fair units
    assert str(yt_density.units) == "g/cm**3"

    # make sure the actual values are the same
    assert_almost_equal(yt_density.d, raw_density)


@requires_module("h5py")
@requires_file(EAGLE_6)
def test_cosmo_dataset_selection():
    ds = load(EAGLE_6)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()



import os
import weakref

import numpy as np

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.funcs import setdefaultattr
from yt.geometry.api import Geometry
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py

from .fields import ChollaFieldInfo


class ChollaGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, index, level, dims):
        super().__init__(id, filename=index.index_filename, index=index)
        self.Parent = None
        self.Children = []
        self.Level = level
        self.ActiveDimensions = dims


class ChollaHierarchy(GridIndex):
    grid = ChollaGrid

    def __init__(self, ds, dataset_type="cholla"):
        self.dataset_type = dataset_type
        self.dataset = weakref.proxy(ds)
        # for now, the index file is the dataset!
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        # float type for the simulation edges and must be float64 now
        self.float_type = np.float64
        super().__init__(ds, dataset_type)

    def _detect_output_fields(self):
        with h5py.File(self.index_filename, mode="r") as h5f:
            self.field_list = [("cholla", k) for k in h5f.keys()]

    def _count_grids(self):
        self.num_grids = 1

    def _parse_index(self):
        self.grid_left_edge[0][:] = self.ds.domain_left_edge[:]
        self.grid_right_edge[0][:] = self.ds.domain_right_edge[:]
        self.grid_dimensions[0][:] = self.ds.domain_dimensions[:]
        self.grid_particle_count[0][0] = 0
        self.grid_levels[0][0] = 0
        self.max_level = 0

    def _populate_grid_objects(self):
        self.grids = np.empty(self.num_grids, dtype="object")
        for i in range(self.num_grids):
            g = self.grid(i, self, self.grid_levels.flat[i], self.grid_dimensions[i])
            g._prepare_grid()
            g._setup_dx()
            self.grids[i] = g


class ChollaDataset(Dataset):
    _load_requirements = ["h5py"]
    _index_class = ChollaHierarchy
    _field_info_class = ChollaFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="cholla",
        storage_filename=None,
        units_override=None,
        unit_system="cgs",
    ):
        self.fluid_types += ("cholla",)
        super().__init__(filename, dataset_type, units_override=units_override)
        self.storage_filename = storage_filename

    def _set_code_unit_attributes(self):
        # This is where quantities are created that represent the various
        # on-disk units.  These are the defaults, but if they are listed
        # in the HDF5 attributes for a file, which is loaded first, then those are
        # used instead.
        #
        if not self.length_unit:
            self.length_unit = self.quan(1.0, "pc")
        if not self.mass_unit:
            self.mass_unit = self.quan(1.0, "Msun")
        if not self.time_unit:
            self.time_unit = self.quan(1000, "yr")
        if not self.velocity_unit:
            self.velocity_unit = self.quan(1.0, "cm/s")
        if not self.magnetic_unit:
            self.magnetic_unit = self.quan(1.0, "gauss")

        for key, unit in self.__class__.default_units.items():
            setdefaultattr(self, key, self.quan(1, unit))

    def _parse_parameter_file(self):
        with h5py.File(self.parameter_filename, mode="r") as h5f:
            attrs = h5f.attrs
            self.parameters = dict(attrs.items())
            self.domain_left_edge = attrs["bounds"][:].astype("=f8")
            self.domain_right_edge = self.domain_left_edge + attrs["domain"][:].astype(
                "=f8"
            )
            self.dimensionality = len(attrs["dims"][:])
            self.domain_dimensions = attrs["dims"][:].astype("=f8")
            self.current_time = attrs["t"][:]
            self._periodicity = tuple(attrs.get("periodicity", (False, False, False)))
            self.gamma = attrs.get("gamma", 5.0 / 3.0)
            if (self.default_species_fields is not None) and "mu" in attrs:
                raise ValueError(
                    'default_species_fields must be None when "mu" is an hdf5 attribute'
                )
            elif "mu" in attrs:
                self.mu = attrs["mu"]
            elif self.default_species_fields is None:
                # other yt-machinery can't handle ds.mu == None, so we simply
                # avoid defining the mu attribute if we don't know its value
                mylog.info(
                    'add the "mu" hdf5 attribute OR use the default_species_fields kwarg '
                    "to compute temperature"
                )
            self.refine_by = 1

            # If header specifies code units, default to those (in CGS)
            length_unit = attrs.get("length_unit", None)
            mass_unit = attrs.get("mass_unit", None)
            time_unit = attrs.get("time_unit", None)
            velocity_unit = attrs.get("velocity_unit", None)
            magnetic_unit = attrs.get("magnetic_unit", None)
            if length_unit:
                self.length_unit = self.quan(length_unit[0], "cm")
            if mass_unit:
                self.mass_unit = self.quan(mass_unit[0], "g")
            if time_unit:
                self.time_unit = self.quan(time_unit[0], "s")
            if velocity_unit:
                self.velocity_unit = self.quan(velocity_unit[0], "cm/s")
            if magnetic_unit:
                self.magnetic_unit = self.quan(magnetic_unit[0], "gauss")

            # this minimalistic implementation fills the requirements for
            # this frontend to run, change it to make it run _correctly_ !
            for key, unit in self.__class__.default_units.items():
                setdefaultattr(self, key, self.quan(1, unit))

        # CHOLLA cannot yet be run as a cosmological simulation
        self.cosmological_simulation = 0
        self.current_redshift = 0.0
        self.omega_lambda = 0.0
        self.omega_matter = 0.0
        self.hubble_constant = 0.0

        # CHOLLA datasets are always unigrid cartesian
        self.geometry = Geometry.CARTESIAN

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        # This accepts a filename or a set of arguments and returns True or
        # False depending on if the file is of the type requested.
        if cls._missing_load_requirements():
            return False

        try:
            fileh = h5py.File(filename, mode="r")
        except OSError:
            return False

        try:
            attrs = fileh.attrs
        except AttributeError:
            return False
        else:
            return (
                "bounds" in attrs
                and "domain" in attrs
                and attrs.get("data_type") != "yt_light_ray"
            )
        finally:
            fileh.close()



import os
import shutil
import tempfile
from unittest import TestCase

import yt
from yt.testing import fake_random_ds


def setup_module():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


class SaveRenderTest(TestCase):
    # This toggles using a temporary directory. Turn off to examine images.
    use_tmpdir = True
    tmpdir = "./"

    def setUp(self):
        if self.use_tmpdir:
            tempfile.mkdtemp()
            self.tmpdir = tempfile.mkdtemp()

    def tearDown(self):
        if self.use_tmpdir:
            shutil.rmtree(self.tmpdir)

    def test_save_render(self):
        ds = fake_random_ds(ndims=32)
        sc = yt.create_scene(ds)

        # make sure it renders if nothing exists, even if render = False
        sc.save(os.path.join(self.tmpdir, "raw.png"), render=False)
        # make sure it re-renders
        sc.save(os.path.join(self.tmpdir, "raw_2.png"), render=True)
        # make sure sigma clip does not re-render
        sc.save(os.path.join(self.tmpdir, "clip_2.png"), sigma_clip=2.0, render=False)
        sc.save(os.path.join(self.tmpdir, "clip_4.png"), sigma_clip=4.0, render=False)

        # save a different format with/without sigma clips
        sc.save(os.path.join(self.tmpdir, "no_clip.jpg"), render=False)
        sc.save(os.path.join(self.tmpdir, "clip_2.jpg"), sigma_clip=2, render=False)



from collections import defaultdict

from numpy.testing import assert_array_almost_equal, assert_equal

import yt
from yt.testing import requires_file, skip

isothermal_h5 = "IsothermalCollapse/snap_505.hdf5"
isothermal_bin = "IsothermalCollapse/snap_505"
snap_33 = "snapshot_033/snap_033.0.hdf5"
tipsy_gal = "TipsyGalaxy/galaxy.00300"
FIRE_m12i = "FIRE_M12i_ref11/snapshot_600.hdf5"

iso_kwargs = {
    "bounding_box": [[-3, 3], [-3, 3], [-3, 3]],
    "unit_base": {
        "UnitLength_in_cm": 5.0e16,
        "UnitMass_in_g": 1.98992e33,
        "UnitVelocity_in_cm_per_s": 46385.190,
    },
}

load_kwargs = defaultdict(dict)
load_kwargs.update(
    {
        isothermal_h5: iso_kwargs,
        isothermal_bin: iso_kwargs,
    }
)

gas_fields_to_particle_fields = {
    "temperature": "Temperature",
    "density": "Density",
    "velocity_x": "particle_velocity_x",
    "velocity_magnitude": "particle_velocity_magnitude",
}


@skip(reason="See https://github.com/yt-project/yt/issues/3909")
@requires_file(isothermal_bin)
@requires_file(isothermal_h5)
@requires_file(snap_33)
@requires_file(tipsy_gal)
@requires_file(FIRE_m12i)
def test_sph_field_semantics():
    for ds_fn in [tipsy_gal, isothermal_h5, isothermal_bin, snap_33, FIRE_m12i]:
        yield sph_fields_validate, ds_fn


def sph_fields_validate(ds_fn):
    ds = yt.load(ds_fn, **(load_kwargs[ds_fn]))
    ad = ds.all_data()
    for gf, pf in gas_fields_to_particle_fields.items():
        gas_field = ad["gas", gf]
        part_field = ad[ds._sph_ptypes[0], pf]

        assert_array_almost_equal(gas_field, part_field)

        npart = ds.particle_type_counts[ds._sph_ptypes[0]]
        err_msg = f"Field {gf} is not the correct shape"
        assert_equal(npart, gas_field.shape[0], err_msg=err_msg)

    dd = ds.r[0.4:0.6, 0.4:0.6, 0.4:0.6]

    for i, ax in enumerate("xyz"):
        dd.set_field_parameter(f"cp_{ax}_vec", yt.YTArray([1, 1, 1]))
        dd.set_field_parameter("axis", i)
    dd.set_field_parameter("omega_baryon", 0.3)

    for f in ds.fields.gas:
        gas_field = dd[f]
        assert f.is_sph_field



import glob
import os
import sys
from collections import defaultdict
from distutils.ccompiler import get_default_compiler
from importlib import resources as importlib_resources

from setuptools import Distribution, setup

# ensure enclosing directory is in PYTHON_PATH to allow importing from setupext.py
if (script_dir := os.path.dirname(__file__)) not in sys.path:
    sys.path.insert(0, script_dir)

from setupext import (
    NUMPY_MACROS,
    check_CPP14_flags,
    check_for_openmp,
    check_for_pyembree,
    create_build_ext,
    get_python_include_dirs,
    get_setup_options,
    install_ccompiler,
)

install_ccompiler()

if os.path.exists("MANIFEST"):
    os.remove("MANIFEST")

with open("README.md") as file:
    long_description = file.read()

CPP14_CONFIG = defaultdict(
    lambda: check_CPP14_flags(["-std=c++14", "-std=c++1y", "-std=gnu++0x"]),
    {"msvc": ["/std:c++14"]},
)
CPP11_CONFIG = defaultdict(lambda: ["-std=c++11"], {"msvc": ["/std:c++11"]})

_COMPILER = get_default_compiler()

omp_args, _ = check_for_openmp()

if os.name == "nt":
    std_libs = []
else:
    std_libs = ["m"]

CPP14_FLAG = CPP14_CONFIG[_COMPILER]
CPP11_FLAG = CPP11_CONFIG[_COMPILER]

FIXED_INTERP = "fixed_interpolator"

cythonize_aliases = {
    "LIB_DIR": "yt/utilities/lib/",
    "LIB_DIR_GEOM": ["yt/utilities/lib/", "yt/geometry/"],
    "LIB_DIR_GEOM_ARTIO": [
        "yt/utilities/lib/",
        "yt/geometry/",
        "yt/frontends/artio/artio_headers/",
    ],
    "STD_LIBS": std_libs,
    "EWAH_LIBS": std_libs
    + [os.path.abspath(importlib_resources.files("ewah_bool_utils"))],
    "OMP_ARGS": omp_args,
    "FIXED_INTERP": FIXED_INTERP,
    "ARTIO_SOURCE": sorted(glob.glob("yt/frontends/artio/artio_headers/*.c")),
    "CPP14_FLAG": CPP14_FLAG,
    "CPP11_FLAG": CPP11_FLAG,
}

lib_exts = [
    "yt/geometry/*.pyx",
    "yt/utilities/cython_fortran_utils.pyx",
    "yt/frontends/ramses/io_utils.pyx",
    "yt/frontends/gamer/cfields.pyx",
    "yt/utilities/lib/cykdtree/kdtree.pyx",
    "yt/utilities/lib/cykdtree/utils.pyx",
    "yt/frontends/artio/_artio_caller.pyx",
    "yt/utilities/lib/*.pyx",
]

embree_libs, embree_aliases = check_for_pyembree(std_libs)
cythonize_aliases.update(embree_aliases)
lib_exts += embree_libs

# This overrides using lib_exts, so it has to happen after lib_exts is fully defined
build_ext, sdist = create_build_ext(lib_exts, cythonize_aliases)


# Force setuptools to consider that there are ext modules, even if empty.
# See https://github.com/yt-project/yt/issues/2922 and
# https://stackoverflow.com/a/62668026/2601223 for the fix.
class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name."""

    def has_ext_modules(self):
        return True


if __name__ == "__main__":
    # Avoid a race condition on fixed_interpolator.o during parallel builds by
    # building it only once and storing it in a static library.
    # See https://github.com/yt-project/yt/issues/4278 and
    # https://github.com/pypa/setuptools/issues/3119#issuecomment-2076922303
    # for the inspiration for this fix.

    # build_clib doesn't add the Python include dirs (for Python.h) by default,
    # as opposed to build_ext, so we need to add them manually.
    clib_include_dirs = get_python_include_dirs()

    # fixed_interpolator.cpp uses Numpy types
    import numpy

    clib_include_dirs.append(numpy.get_include())

    fixed_interp_lib = (
        FIXED_INTERP,
        {
            "sources": ["yt/utilities/lib/fixed_interpolator.cpp"],
            "include_dirs": clib_include_dirs,
            "define_macros": NUMPY_MACROS,
        },
    )

    setup(
        cmdclass={"sdist": sdist, "build_ext": build_ext},
        distclass=BinaryDistribution,
        libraries=[fixed_interp_lib],
        ext_modules=[],  # !!! We override this inside build_ext above
        options=get_setup_options(),
    )



import glob
import os
from collections import defaultdict

import numpy as np

from yt.data_objects.static_output import ParticleDataset, ParticleFile
from yt.frontends.gadget.data_structures import _fix_unit_ordering
from yt.funcs import only_on_root, setdefaultattr
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.utilities.exceptions import YTException
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py

from .fields import OWLSSubfindFieldInfo


class OWLSSubfindParticleIndex(ParticleIndex):
    chunksize = -1

    def __init__(self, ds, dataset_type):
        super().__init__(ds, dataset_type)

    def _calculate_particle_index_starts(self):
        # Halo indices are not saved in the file, so we must count by hand.
        # File 0 has halos 0 to N_0 - 1, file 1 has halos N_0 to N_0 + N_1 - 1, etc.
        particle_count = defaultdict(int)
        offset_count = 0
        for data_file in self.data_files:
            data_file.index_start = {
                ptype: particle_count[ptype] for ptype in data_file.total_particles
            }
            data_file.offset_start = offset_count
            for ptype in data_file.total_particles:
                particle_count[ptype] += data_file.total_particles[ptype]
            offset_count += data_file.total_offset

    def _calculate_file_offset_map(self):
        # After the FOF  is performed, a load-balancing step redistributes halos
        # and then writes more fields.  Here, for each file, we create a list of
        # files which contain the rest of the redistributed particles.
        ifof = np.array(
            [data_file.total_particles["FOF"] for data_file in self.data_files]
        )
        isub = np.array([data_file.total_offset for data_file in self.data_files])
        subend = isub.cumsum()
        fofend = ifof.cumsum()
        istart = np.digitize(fofend - ifof, subend - isub) - 1
        iend = np.clip(np.digitize(fofend, subend), 0, ifof.size - 2)
        for i, data_file in enumerate(self.data_files):
            data_file.offset_files = self.data_files[istart[i] : iend[i] + 1]

    def _detect_output_fields(self):
        # TODO: Add additional fields
        self._calculate_particle_index_starts()
        self._calculate_file_offset_map()
        dsl = []
        units = {}
        for dom in self.data_files:
            fl, _units = self.io._identify_fields(dom)
            units.update(_units)
            for f in fl:
                if f not in dsl:
                    dsl.append(f)
        self.field_list = dsl
        ds = self.dataset
        ds.particle_types = tuple({pt for pt, ds in dsl})
        # This is an attribute that means these particle types *actually*
        # exist.  As in, they are real, in the dataset.
        ds.field_units.update(units)
        ds.particle_types_raw = ds.particle_types


class OWLSSubfindHDF5File(ParticleFile):
    def __init__(self, ds, io, filename, file_id, bounds):
        super().__init__(ds, io, filename, file_id, bounds)
        with h5py.File(filename, mode="r") as f:
            self.header = {field: f.attrs[field] for field in f.attrs.keys()}


class OWLSSubfindDataset(ParticleDataset):
    _load_requirements = ["h5py"]
    _index_class = OWLSSubfindParticleIndex
    _file_class = OWLSSubfindHDF5File
    _field_info_class = OWLSSubfindFieldInfo
    _suffix = ".hdf5"

    def __init__(
        self,
        filename,
        dataset_type="subfind_hdf5",
        index_order=None,
        index_filename=None,
        units_override=None,
        unit_system="cgs",
    ):
        super().__init__(
            filename,
            dataset_type,
            index_order=index_order,
            index_filename=index_filename,
            units_override=units_override,
            unit_system=unit_system,
        )

    def _parse_parameter_file(self):
        handle = h5py.File(self.parameter_filename, mode="r")
        hvals = {}
        hvals.update((str(k), v) for k, v in handle["/Header"].attrs.items())
        hvals["NumFiles"] = hvals["NumFilesPerSnapshot"]
        hvals["Massarr"] = hvals["MassTable"]

        self.dimensionality = 3
        self.refine_by = 2

        # Set standard values
        if "Time_GYR" in hvals:
            self.current_time = self.quan(hvals["Time_GYR"], "Gyr")
        self.domain_left_edge = np.zeros(3, "float64")
        self.domain_right_edge = np.ones(3, "float64") * hvals["BoxSize"]
        self.domain_dimensions = np.ones(3, "int32")
        self.cosmological_simulation = 1
        self._periodicity = (True, True, True)
        self.current_redshift = hvals["Redshift"]
        self.omega_lambda = hvals["OmegaLambda"]
        self.omega_matter = hvals["Omega0"]
        self.hubble_constant = hvals["HubbleParam"]
        self.parameters = hvals
        prefix = os.path.abspath(
            os.path.join(
                os.path.dirname(self.parameter_filename),
                os.path.basename(self.parameter_filename).split(".", 1)[0],
            )
        )

        suffix = self.parameter_filename.rsplit(".", 1)[-1]
        self.filename_template = f"{prefix}.%(num)i.{suffix}"
        self.file_count = len(glob.glob(prefix + "*" + self._suffix))
        if self.file_count == 0:
            raise YTException(message="No data files found.", ds=self)
        self.particle_types = ("FOF", "SUBFIND")
        self.particle_types_raw = ("FOF", "SUBFIND")

        # To avoid having to open files twice
        self._unit_base = {}
        self._unit_base.update((str(k), v) for k, v in handle["/Units"].attrs.items())
        handle.close()

    def _set_code_unit_attributes(self):
        # Set a sane default for cosmological simulations.
        if self._unit_base is None and self.cosmological_simulation == 1:
            only_on_root(mylog.info, "Assuming length units are in Mpc/h (comoving)")
            self._unit_base = {"length": (1.0, "Mpccm/h")}
        # The other same defaults we will use from the standard Gadget
        # defaults.
        unit_base = self._unit_base or {}

        if "length" in unit_base:
            length_unit = unit_base["length"]
        elif "UnitLength_in_cm" in unit_base:
            if self.cosmological_simulation == 0:
                length_unit = (unit_base["UnitLength_in_cm"], "cm")
            else:
                length_unit = (unit_base["UnitLength_in_cm"], "cmcm/h")
        else:
            raise RuntimeError
        length_unit = _fix_unit_ordering(length_unit)
        setdefaultattr(self, "length_unit", self.quan(length_unit[0], length_unit[1]))

        if "velocity" in unit_base:
            velocity_unit = unit_base["velocity"]
        elif "UnitVelocity_in_cm_per_s" in unit_base:
            velocity_unit = (unit_base["UnitVelocity_in_cm_per_s"], "cm/s")
        else:
            velocity_unit = (1e5, "cm/s  * sqrt(a)")
        velocity_unit = _fix_unit_ordering(velocity_unit)
        setdefaultattr(
            self, "velocity_unit", self.quan(velocity_unit[0], velocity_unit[1])
        )

        # We set hubble_constant = 1.0 for non-cosmology, so this is safe.
        # Default to 1e10 Msun/h if mass is not specified.
        if "mass" in unit_base:
            mass_unit = unit_base["mass"]
        elif "UnitMass_in_g" in unit_base:
            if self.cosmological_simulation == 0:
                mass_unit = (unit_base["UnitMass_in_g"], "g")
            else:
                mass_unit = (unit_base["UnitMass_in_g"], "g/h")
        else:
            # Sane default
            mass_unit = (1.0, "1e10*Msun/h")
        mass_unit = _fix_unit_ordering(mass_unit)
        setdefaultattr(self, "mass_unit", self.quan(mass_unit[0], mass_unit[1]))

        if "time" in unit_base:
            time_unit = unit_base["time"]
        elif "UnitTime_in_s" in unit_base:
            time_unit = (unit_base["UnitTime_in_s"], "s")
        else:
            tu = (self.length_unit / self.velocity_unit).to("yr/h")
            time_unit = (tu.d, tu.units)
        setdefaultattr(self, "time_unit", self.quan(time_unit[0], time_unit[1]))

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        need_groups = ["Constants", "Header", "Parameters", "Units", "FOF"]
        valid = True
        try:
            fh = h5py.File(filename, mode="r")
            valid = all(ng in fh["/"] for ng in need_groups)
            fh.close()
        except Exception:
            valid = False
        return valid



import numpy as np

from yt.data_objects.index_subobjects.unstructured_mesh import SemiStructuredMesh
from yt.funcs import mylog
from yt.units._numpy_wrapper_functions import uconcatenate, uvstack
from yt.units.yt_array import YTArray
from yt.utilities.lib.pixelization_routines import (
    interpolate_sph_grid_gather,
    normalization_2d_utility,
    pixelize_cartesian,
    pixelize_cartesian_nodal,
    pixelize_element_mesh,
    pixelize_element_mesh_line,
    pixelize_off_axis_cartesian,
    pixelize_sph_kernel_cutting,
    pixelize_sph_kernel_projection,
    pixelize_sph_kernel_slice,
)
from yt.utilities.math_utils import compute_stddev_image
from yt.utilities.nodal_data_utils import get_nodal_data

from .coordinate_handler import (
    CoordinateHandler,
    _get_coord_fields,
    _get_vert_fields,
    cartesian_to_cylindrical,
    cylindrical_to_cartesian,
)


def _sample_ray(ray, npoints, field):
    """
    Private function that uses a ray object for calculating the field values
    that will be the y-axis values in a LinePlot object.

    Parameters
    ----------
    ray : YTOrthoRay, YTRay, or LightRay
        Ray object from which to sample field values
    npoints : int
        The number of points to sample
    field : str or field tuple
        The name of the field to sample
    """
    start_point = ray.start_point
    end_point = ray.end_point
    sample_dr = (end_point - start_point) / (npoints - 1)
    sample_points = [np.arange(npoints) * sample_dr[i] for i in range(3)]
    sample_points = uvstack(sample_points).T + start_point
    ray_coordinates = uvstack([ray["index", d] for d in "xyz"]).T
    ray_dds = uvstack([ray["index", f"d{d}"] for d in "xyz"]).T
    ray_field = ray[field]
    field_values = ray.ds.arr(np.zeros(npoints), ray_field.units)
    for i, sample_point in enumerate(sample_points):
        ray_contains = (sample_point >= (ray_coordinates - ray_dds / 2)) & (
            sample_point <= (ray_coordinates + ray_dds / 2)
        )
        ray_contains = ray_contains.all(axis=-1)
        # use argmax to find the first nonzero index, sometimes there
        # are two indices if the sampling point happens to fall exactly at
        # a cell boundary
        field_values[i] = ray_field[np.argmax(ray_contains)]
    dr = np.sqrt((sample_dr**2).sum())
    x = np.arange(npoints) / (npoints - 1) * (dr * npoints)
    return x, field_values


def all_data(data, ptype, fields, kdtree=False):
    field_data = {}
    fields = set(fields)
    for field in fields:
        field_data[field] = []

    for chunk in data.all_data().chunks([], "io"):
        for field in fields:
            field_data[field].append(chunk[ptype, field].in_base("code"))

    for field in fields:
        field_data[field] = uconcatenate(field_data[field])

    if kdtree is True:
        kdtree = data.index.kdtree
        for field in fields:
            if len(field_data[field].shape) == 1:
                field_data[field] = field_data[field][kdtree.idx]
            else:
                field_data[field] = field_data[field][kdtree.idx, :]

    return field_data


class CartesianCoordinateHandler(CoordinateHandler):
    name = "cartesian"
    _default_axis_order = ("x", "y", "z")

    def setup_fields(self, registry):
        for axi, ax in enumerate(self.axis_order):
            f1, f2 = _get_coord_fields(axi)
            registry.add_field(
                ("index", f"d{ax}"),
                sampling_type="cell",
                function=f1,
                display_field=False,
                units="code_length",
            )

            registry.add_field(
                ("index", f"path_element_{ax}"),
                sampling_type="cell",
                function=f1,
                display_field=False,
                units="code_length",
            )

            registry.add_field(
                ("index", f"{ax}"),
                sampling_type="cell",
                function=f2,
                display_field=False,
                units="code_length",
            )

            f3 = _get_vert_fields(axi)
            registry.add_field(
                ("index", f"vertex_{ax}"),
                sampling_type="cell",
                function=f3,
                display_field=False,
                units="code_length",
            )

        self._register_volume(registry)
        self._check_fields(registry)

    def _register_volume(self, registry):
        def _cell_volume(field, data):
            rv = data["index", "dx"].copy(order="K")
            rv *= data["index", "dy"]
            rv *= data["index", "dz"]
            return rv

        registry.add_field(
            ("index", "cell_volume"),
            sampling_type="cell",
            function=_cell_volume,
            display_field=False,
            units="code_length**3",
        )
        registry.alias(("index", "volume"), ("index", "cell_volume"))

    def _check_fields(self, registry):
        registry.check_derived_fields(
            [
                ("index", "dx"),
                ("index", "dy"),
                ("index", "dz"),
                ("index", "x"),
                ("index", "y"),
                ("index", "z"),
                ("index", "cell_volume"),
            ]
        )

    def pixelize(
        self,
        dimension,
        data_source,
        field,
        bounds,
        size,
        antialias=True,
        periodic=True,
        *,
        return_mask=False,
    ):
        """
        Method for pixelizing datasets in preparation for
        two-dimensional image plots. Relies on several sampling
        routines written in cython
        """
        index = data_source.ds.index
        if hasattr(index, "meshes") and not isinstance(
            index.meshes[0], SemiStructuredMesh
        ):
            ftype, fname = field
            if ftype == "all":
                mesh_id = 0
                indices = np.concatenate(
                    [mesh.connectivity_indices for mesh in index.mesh_union]
                )
            else:
                mesh_id = int(ftype[-1]) - 1
                indices = index.meshes[mesh_id].connectivity_indices

            coords = index.meshes[mesh_id].connectivity_coords
            offset = index.meshes[mesh_id]._index_offset
            ad = data_source.ds.all_data()
            field_data = ad[field]
            buff_size = size[0:dimension] + (1,) + size[dimension:]

            ax = data_source.axis
            xax = self.x_axis[ax]
            yax = self.y_axis[ax]
            c = np.float64(data_source.center[dimension].d)

            extents = np.zeros((3, 2))
            extents[ax] = np.array([c, c])
            extents[xax] = bounds[0:2]
            extents[yax] = bounds[2:4]

            # if this is an element field, promote to 2D here
            if len(field_data.shape) == 1:
                field_data = np.expand_dims(field_data, 1)
            # if this is a higher-order element, we demote to 1st order
            # here, for now.
            elif field_data.shape[1] == 27:
                # hexahedral
                mylog.warning(
                    "High order elements not yet supported, dropping to 1st order."
                )
                field_data = field_data[:, 0:8]
                indices = indices[:, 0:8]

            buff, mask = pixelize_element_mesh(
                coords,
                indices,
                buff_size,
                field_data,
                extents,
                index_offset=offset,
                return_mask=True,
            )

            buff = np.squeeze(np.transpose(buff, (yax, xax, ax)))
            mask = np.squeeze(np.transpose(mask, (yax, xax, ax)))

        elif self.axis_id.get(dimension, dimension) is not None:
            buff, mask = self._ortho_pixelize(
                data_source, field, bounds, size, antialias, dimension, periodic
            )
        else:
            buff, mask = self._oblique_pixelize(
                data_source, field, bounds, size, antialias
            )

        if return_mask:
            assert mask is None or mask.dtype == bool
            return buff, mask
        else:
            return buff

    def pixelize_line(self, field, start_point, end_point, npoints):
        """
        Method for sampling datasets along a line in preparation for
        one-dimensional line plots. For UnstructuredMesh, relies on a
        sampling routine written in cython
        """
        if npoints < 2:
            raise ValueError(
                "Must have at least two sample points in order to draw a line plot."
            )
        index = self.ds.index
        if hasattr(index, "meshes") and not isinstance(
            index.meshes[0], SemiStructuredMesh
        ):
            ftype, fname = field
            if ftype == "all":
                mesh_id = 0
                indices = np.concatenate(
                    [mesh.connectivity_indices for mesh in index.mesh_union]
                )
            else:
                mesh_id = int(ftype[-1]) - 1
                indices = index.meshes[mesh_id].connectivity_indices

            coords = index.meshes[mesh_id].connectivity_coords
            if coords.shape[1] != end_point.size != start_point.size:
                raise ValueError(
                    "The coordinate dimension doesn't match the "
                    "start and end point dimensions."
                )

            offset = index.meshes[mesh_id]._index_offset
            ad = self.ds.all_data()
            field_data = ad[field]

            # if this is an element field, promote to 2D here
            if len(field_data.shape) == 1:
                field_data = np.expand_dims(field_data, 1)
            # if this is a higher-order element, we demote to 1st order
            # here, for now.
            elif field_data.shape[1] == 27:
                # hexahedral
                mylog.warning(
                    "High order elements not yet supported, dropping to 1st order."
                )
                field_data = field_data[:, 0:8]
                indices = indices[:, 0:8]

            arc_length, plot_values = pixelize_element_mesh_line(
                coords,
                indices,
                start_point,
                end_point,
                npoints,
                field_data,
                index_offset=offset,
            )
            arc_length = YTArray(arc_length, start_point.units)
            plot_values = YTArray(plot_values, field_data.units)
        else:
            ray = self.ds.ray(start_point, end_point)
            arc_length, plot_values = _sample_ray(ray, npoints, field)
        return arc_length, plot_values

    def _ortho_pixelize(
        self, data_source, field, bounds, size, antialias, dim, periodic
    ):
        from yt.data_objects.construction_data_containers import YTParticleProj
        from yt.data_objects.selection_objects.slices import YTSlice
        from yt.frontends.sph.data_structures import ParticleDataset
        from yt.frontends.stream.data_structures import StreamParticlesDataset

        # We should be using fcoords
        field = data_source._determine_fields(field)[0]
        finfo = data_source.ds.field_info[field]
        # some coordinate handlers use only projection-plane periods,
        # others need all box periods.
        period2 = self.period[:2].copy()  # dummy here
        period2[0] = self.period[self.x_axis[dim]]
        period2[1] = self.period[self.y_axis[dim]]
        period3 = self.period[:].copy()  # dummy here
        period3[0] = self.period[self.x_axis[dim]]
        period3[1] = self.period[self.y_axis[dim]]
        zax = list({0, 1, 2} - {self.x_axis[dim], self.y_axis[dim]})[0]
        period3[2] = self.period[zax]
        if hasattr(period2, "in_units"):
            period2 = period2.in_units("code_length").d
        if hasattr(period3, "in_units"):
            period3 = period3.in_units("code_length").d

        buff = np.full((size[1], size[0]), np.nan, dtype="float64")
        particle_datasets = (ParticleDataset, StreamParticlesDataset)
        is_sph_field = finfo.is_sph_field

        finfo = self.ds._get_field_info(field)
        if np.any(finfo.nodal_flag):
            nodal_data = get_nodal_data(data_source, field)
            coord = data_source.coord.d
            mask = pixelize_cartesian_nodal(
                buff,
                data_source["px"],
                data_source["py"],
                data_source["pz"],
                data_source["pdx"],
                data_source["pdy"],
                data_source["pdz"],
                nodal_data,
                coord,
                bounds,
                int(antialias),
                period2,
                int(periodic),
                return_mask=True,
            )
        elif isinstance(data_source.ds, particle_datasets) and is_sph_field:
            # SPH handling
            ptype = field[0]
            if ptype == "gas":
                ptype = data_source.ds._sph_ptypes[0]
            px_name = self.axis_name[self.x_axis[dim]]
            py_name = self.axis_name[self.y_axis[dim]]
            # need z coordinates for depth,
            # but name isn't saved in the handler -> use the 'other one'
            pz_name = list(set(self.axis_order) - {px_name, py_name})[0]

            # ignore default True periodic argument
            # (not actually supplied by a call from
            # FixedResolutionBuffer), and use the dataset periodicity
            # instead
            xa = self.x_axis[dim]
            ya = self.y_axis[dim]
            # axorder = data_source.ds.coordinates.axis_order
            za = list({0, 1, 2} - {xa, ya})[0]
            ds_periodic = data_source.ds.periodicity
            _periodic = np.array(ds_periodic)
            _periodic[0] = ds_periodic[xa]
            _periodic[1] = ds_periodic[ya]
            _periodic[2] = ds_periodic[za]
            ounits = data_source.ds.field_info[field].output_units
            bnds = data_source.ds.arr(bounds, "code_length").tolist()
            kernel_name = None
            if hasattr(data_source.ds, "kernel_name"):
                kernel_name = data_source.ds.kernel_name
            if kernel_name is None:
                kernel_name = "cubic"

            if isinstance(data_source, YTParticleProj):  # projection
                weight = data_source.weight_field
                moment = data_source.moment
                le, re = data_source.data_source.get_bbox()
                # If we're not periodic, we need to clip to the boundary edges
                # or we get errors about extending off the edge of the region.
                # (depth/z range is handled by region setting)
                if not self.ds.periodicity[xa]:
                    le[xa] = max(bounds[0], self.ds.domain_left_edge[xa])
                    re[xa] = min(bounds[1], self.ds.domain_right_edge[xa])
                else:
                    le[xa] = bounds[0]
                    re[xa] = bounds[1]
                if not self.ds.periodicity[ya]:
                    le[ya] = max(bounds[2], self.ds.domain_left_edge[ya])
                    re[ya] = min(bounds[3], self.ds.domain_right_edge[ya])
                else:
                    le[ya] = bounds[2]
                    re[ya] = bounds[3]
                # We actually need to clip these
                proj_reg = data_source.ds.region(
                    left_edge=le,
                    right_edge=re,
                    center=data_source.center,
                    data_source=data_source.data_source,
                )
                proj_reg.set_field_parameter("axis", data_source.axis)
                # need some z bounds for SPH projection
                # -> use source bounds
                bnds3 = bnds + [le[za], re[za]]

                buff = np.zeros(size, dtype="float64")
                mask_uint8 = np.zeros_like(buff, dtype="uint8")
                if weight is None:
                    for chunk in proj_reg.chunks([], "io"):
                        data_source._initialize_projected_units([field], chunk)
                        pixelize_sph_kernel_projection(
                            buff,
                            mask_uint8,
                            chunk[ptype, px_name].to("code_length"),
                            chunk[ptype, py_name].to("code_length"),
                            chunk[ptype, pz_name].to("code_length"),
                            chunk[ptype, "smoothing_length"].to("code_length"),
                            chunk[ptype, "mass"].to("code_mass"),
                            chunk[ptype, "density"].to("code_density"),
                            chunk[field].in_units(ounits),
                            bnds3,
                            _check_period=_periodic.astype("int"),
                            period=period3,
                            kernel_name=kernel_name,
                        )
                    # We use code length here, but to get the path length right
                    # we need to multiply by the conversion factor between
                    # code length and the unit system's length unit
                    default_path_length_unit = data_source.ds.unit_system["length"]
                    dl_conv = data_source.ds.quan(1.0, "code_length").to(
                        default_path_length_unit
                    )
                    buff *= dl_conv.v
                # if there is a weight field, take two projections:
                # one of field*weight, the other of just weight, and divide them
                else:
                    weight_buff = np.zeros(size, dtype="float64")
                    buff = np.zeros(size, dtype="float64")
                    mask_uint8 = np.zeros_like(buff, dtype="uint8")
                    wounits = data_source.ds.field_info[weight].output_units
                    for chunk in proj_reg.chunks([], "io"):
                        data_source._initialize_projected_units([field], chunk)
                        data_source._initialize_projected_units([weight], chunk)
                        pixelize_sph_kernel_projection(
                            buff,
                            mask_uint8,
                            chunk[ptype, px_name].to("code_length"),
                            chunk[ptype, py_name].to("code_length"),
                            chunk[ptype, pz_name].to("code_length"),
                            chunk[ptype, "smoothing_length"].to("code_length"),
                            chunk[ptype, "mass"].to("code_mass"),
                            chunk[ptype, "density"].to("code_density"),
                            chunk[field].in_units(ounits),
                            bnds3,
                            _check_period=_periodic.astype("int"),
                            period=period3,
                            weight_field=chunk[weight].in_units(wounits),
                            kernel_name=kernel_name,
                        )
                    mylog.info(
                        "Making a fixed resolution buffer of (%s) %d by %d",
                        weight,
                        size[0],
                        size[1],
                    )
                    for chunk in proj_reg.chunks([], "io"):
                        data_source._initialize_projected_units([weight], chunk)
                        pixelize_sph_kernel_projection(
                            weight_buff,
                            mask_uint8,
                            chunk[ptype, px_name].to("code_length"),
                            chunk[ptype, py_name].to("code_length"),
                            chunk[ptype, pz_name].to("code_length"),
                            chunk[ptype, "smoothing_length"].to("code_length"),
                            chunk[ptype, "mass"].to("code_mass"),
                            chunk[ptype, "density"].to("code_density"),
                            chunk[weight].in_units(wounits),
                            bnds3,
                            _check_period=_periodic.astype("int"),
                            period=period3,
                            kernel_name=kernel_name,
                        )
                    normalization_2d_utility(buff, weight_buff)
                    if moment == 2:
                        buff2 = np.zeros(size, dtype="float64")
                        for chunk in proj_reg.chunks([], "io"):
                            data_source._initialize_projected_units([field], chunk)
                            data_source._initialize_projected_units([weight], chunk)
                            pixelize_sph_kernel_projection(
                                buff2,
                                mask_uint8,
                                chunk[ptype, px_name].to("code_length"),
                                chunk[ptype, py_name].to("code_length"),
                                chunk[ptype, pz_name].to("code_length"),
                                chunk[ptype, "smoothing_length"].to("code_length"),
                                chunk[ptype, "mass"].to("code_mass"),
                                chunk[ptype, "density"].to("code_density"),
                                chunk[field].in_units(ounits) ** 2,
                                bnds3,
                                _check_period=_periodic.astype("int"),
                                period=period3,
                                weight_field=chunk[weight].in_units(wounits),
                                kernel_name=kernel_name,
                            )
                        normalization_2d_utility(buff2, weight_buff)
                        buff = compute_stddev_image(buff2, buff)
                mask = mask_uint8.astype("bool")
            elif isinstance(data_source, YTSlice):
                smoothing_style = getattr(self.ds, "sph_smoothing_style", "scatter")
                normalize = getattr(self.ds, "use_sph_normalization", True)

                if smoothing_style == "scatter":
                    buff = np.zeros(size, dtype="float64")
                    mask_uint8 = np.zeros_like(buff, dtype="uint8")
                    if normalize:
                        buff_den = np.zeros(size, dtype="float64")

                    for chunk in data_source.chunks([], "io"):
                        hsmlname = "smoothing_length"
                        pixelize_sph_kernel_slice(
                            buff,
                            mask_uint8,
                            chunk[ptype, px_name].to("code_length").v,
                            chunk[ptype, py_name].to("code_length").v,
                            chunk[ptype, pz_name].to("code_length").v,
                            chunk[ptype, hsmlname].to("code_length").v,
                            chunk[ptype, "mass"].to("code_mass").v,
                            chunk[ptype, "density"].to("code_density").v,
                            chunk[field].in_units(ounits).v,
                            bnds,
                            data_source.coord.to("code_length").v,
                            _check_period=_periodic.astype("int"),
                            period=period3,
                            kernel_name=kernel_name,
                        )
                        if normalize:
                            pixelize_sph_kernel_slice(
                                buff_den,
                                mask_uint8,
                                chunk[ptype, px_name].to("code_length").v,
                                chunk[ptype, py_name].to("code_length").v,
                                chunk[ptype, pz_name].to("code_length").v,
                                chunk[ptype, hsmlname].to("code_length").v,
                                chunk[ptype, "mass"].to("code_mass").v,
                                chunk[ptype, "density"].to("code_density").v,
                                np.ones(chunk[ptype, "density"].shape[0]),
                                bnds,
                                data_source.coord.to("code_length").v,
                                _check_period=_periodic.astype("int"),
                                period=period3,
                                kernel_name=kernel_name,
                            )

                    if normalize:
                        normalization_2d_utility(buff, buff_den)

                    mask = mask_uint8.astype("bool", copy=False)

                if smoothing_style == "gather":
                    # Here we find out which axis are going to be the "x" and
                    # "y" axis for the actual visualisation and then we set the
                    # buffer size and bounds to match. The z axis of the plot
                    # is the axis we slice over and the buffer will be of size 1
                    # in that dimension
                    x, y, z = self.x_axis[dim], self.y_axis[dim], dim

                    buff_size = np.zeros(3, dtype="int64")
                    buff_size[x] = size[0]
                    buff_size[y] = size[1]
                    buff_size[z] = 1

                    buff_bounds = np.zeros(6, dtype="float64")
                    buff_bounds[2 * x : 2 * x + 2] = bounds[0:2]
                    buff_bounds[2 * y : 2 * y + 2] = bounds[2:4]
                    buff_bounds[2 * z] = data_source.coord
                    buff_bounds[2 * z + 1] = data_source.coord

                    # then we do the interpolation
                    buff_temp = np.zeros(buff_size, dtype="float64")

                    fields_to_get = [
                        "particle_position",
                        "density",
                        "mass",
                        "smoothing_length",
                        field[1],
                    ]
                    all_fields = all_data(self.ds, ptype, fields_to_get, kdtree=True)

                    num_neighbors = getattr(self.ds, "num_neighbors", 32)
                    mask_temp = interpolate_sph_grid_gather(
                        buff_temp,
                        all_fields["particle_position"].to("code_length"),
                        buff_bounds,
                        all_fields["smoothing_length"].to("code_length"),
                        all_fields["mass"].to("code_mass"),
                        all_fields["density"].to("code_density"),
                        all_fields[field[1]].in_units(ounits),
                        self.ds.index.kdtree,
                        num_neigh=num_neighbors,
                        use_normalization=normalize,
                        return_mask=True,
                    )

                    # We swap the axes back so the axis which was sliced over
                    # is the last axis, as this is the "z" axis of the plots.
                    if z != 2:
                        buff_temp = buff_temp.swapaxes(2, z)
                        mask_temp = mask_temp.swapaxes(2, z)
                        if x == 2:
                            x = z
                        else:
                            y = z

                    buff = buff_temp[:, :, 0]
                    mask = mask_temp[:, :, 0]

                    # Then we just transpose if the buffer x and y are
                    # different than the plot x and y
                    if y < x:
                        buff = buff.transpose()
                        mask = mask.transpose()
            else:
                raise NotImplementedError(
                    "A pixelization routine has not been implemented for "
                    f"{type(data_source)} data objects"
                )
            buff = buff.transpose()
            mask = mask.transpose()
        else:
            mask = pixelize_cartesian(
                buff,
                data_source["px"],
                data_source["py"],
                data_source["pdx"],
                data_source["pdy"],
                data_source[field],
                bounds,
                int(antialias),
                period2,
                int(periodic),
                return_mask=True,
            )
        assert mask is None or mask.dtype == bool
        return buff, mask

    def _oblique_pixelize(self, data_source, field, bounds, size, antialias):
        from yt.data_objects.selection_objects.slices import YTCuttingPlane
        from yt.frontends.sph.data_structures import ParticleDataset
        from yt.frontends.stream.data_structures import StreamParticlesDataset
        from yt.frontends.ytdata.data_structures import YTSpatialPlotDataset

        # Determine what sort of data we're dealing with
        # -> what backend to use
        # copied from the _ortho_pixelize method
        field = data_source._determine_fields(field)[0]
        _finfo = data_source.ds.field_info[field]
        is_sph_field = _finfo.is_sph_field
        particle_datasets = (ParticleDataset, StreamParticlesDataset)
        # finfo = self.ds._get_field_info(field)

        # SPH data
        # only for slices: a function in off_axis_projection.py
        # handles projections
        if (
            isinstance(data_source.ds, particle_datasets)
            and is_sph_field
            and isinstance(data_source, YTCuttingPlane)
        ):
            normalize = getattr(self.ds, "use_sph_normalization", True)
            le = data_source.ds.domain_left_edge.to("code_length")
            re = data_source.ds.domain_right_edge.to("code_length")
            boxbounds = np.array([le[0], re[0], le[1], re[1], le[2], re[2]])
            periodic = data_source.ds.periodicity
            ptype = field[0]
            if ptype == "gas":
                ptype = data_source.ds._sph_ptypes[0]
            axorder = data_source.ds.coordinates.axis_order
            ounits = data_source.ds.field_info[field].output_units
            # input bounds are in code length units already
            widthxy = np.array((bounds[1] - bounds[0], bounds[3] - bounds[2]))
            kernel_name = None
            if hasattr(data_source.ds, "kernel_name"):
                kernel_name = data_source.ds.kernel_name
            if kernel_name is None:
                kernel_name = "cubic"
            # data_source should be a YTCuttingPlane object
            # dimensionless unyt normal/north
            # -> numpy array cython can deal with
            normal_vector = data_source.normal.v
            north_vector = data_source._y_vec.v
            center = data_source.center.to("code_length")

            buff = np.zeros(size, dtype="float64")
            mask_uint8 = np.zeros_like(buff, dtype="uint8")
            if normalize:
                buff_den = np.zeros(size, dtype="float64")

            for chunk in data_source.chunks([], "io"):
                pixelize_sph_kernel_cutting(
                    buff,
                    mask_uint8,
                    chunk[ptype, axorder[0]].to("code_length").v,
                    chunk[ptype, axorder[1]].to("code_length").v,
                    chunk[ptype, axorder[2]].to("code_length").v,
                    chunk[ptype, "smoothing_length"].to("code_length").v,
                    chunk[ptype, "mass"].to("code_mass"),
                    chunk[ptype, "density"].to("code_density"),
                    chunk[field].in_units(ounits),
                    center,
                    widthxy,
                    normal_vector,
                    north_vector,
                    boxbounds,
                    periodic,
                    kernel_name=kernel_name,
                    check_period=1,
                )
                if normalize:
                    pixelize_sph_kernel_cutting(
                        buff_den,
                        mask_uint8,
                        chunk[ptype, axorder[0]].to("code_length"),
                        chunk[ptype, axorder[1]].to("code_length"),
                        chunk[ptype, axorder[2]].to("code_length"),
                        chunk[ptype, "smoothing_length"].to("code_length"),
                        chunk[ptype, "mass"].to("code_mass"),
                        chunk[ptype, "density"].to("code_density"),
                        np.ones(chunk[ptype, "density"].shape[0]),
                        center,
                        widthxy,
                        normal_vector,
                        north_vector,
                        boxbounds,
                        periodic,
                        kernel_name=kernel_name,
                        check_period=1,
                    )

            if normalize:
                normalization_2d_utility(buff, buff_den)

            mask = mask_uint8.astype("bool", copy=False)
            # swap axes for image plotting
            mask = mask.swapaxes(0, 1)
            buff = buff.swapaxes(0, 1)

        # whatever other data this code could handle before the
        # SPH option was added
        else:
            indices = np.argsort(data_source["pdx"])[::-1].astype("int64", copy=False)
            buff = np.full((size[1], size[0]), np.nan, dtype="float64")
            ftype = "index"
            if isinstance(data_source.ds, YTSpatialPlotDataset):
                ftype = "gas"
            mask = pixelize_off_axis_cartesian(
                buff,
                data_source[ftype, "x"],
                data_source[ftype, "y"],
                data_source[ftype, "z"],
                data_source["px"],
                data_source["py"],
                data_source["pdx"],
                data_source["pdy"],
                data_source["pdz"],
                data_source.center,
                data_source._inv_mat,
                indices,
                data_source[field],
                bounds,
            )
        return buff, mask

    def convert_from_cartesian(self, coord):
        return coord

    def convert_to_cartesian(self, coord):
        return coord

    def convert_to_cylindrical(self, coord):
        center = self.ds.domain_center
        return cartesian_to_cylindrical(coord, center)

    def convert_from_cylindrical(self, coord):
        center = self.ds.domain_center
        return cylindrical_to_cartesian(coord, center)

    def convert_to_spherical(self, coord):
        raise NotImplementedError

    def convert_from_spherical(self, coord):
        raise NotImplementedError

    _x_pairs = (("x", "y"), ("y", "z"), ("z", "x"))
    _y_pairs = (("x", "z"), ("y", "x"), ("z", "y"))

    @property
    def period(self):
        return self.ds.domain_width



import pytest

import yt
from yt.testing import fake_amr_ds, requires_module_pytest as requires_module


class TestGeoTransform:
    # Cartopy require pykdtree *or* scipy to enable the projections
    # we test here. We require scipy for simplicity because it offers
    # better support for various platforms via PyPI at the time of writing.abs

    # the following projections are skipped (reason)
    # - UTM (requires additional arguments)
    # - OSNI (avoid crashes, see https://github.com/SciTools/cartopy/issues/1177)

    @classmethod
    def setup_class(cls):
        cls.ds = fake_amr_ds(geometry="geographic")

    @requires_module("cartopy", "scipy")
    @pytest.mark.parametrize(
        "transform",
        [
            "PlateCarree",
            "LambertConformal",
            "LambertCylindrical",
            "Mercator",
            "Miller",
            "Mollweide",
            "Orthographic",
            "Robinson",
            "Stereographic",
            "TransverseMercator",
            "InterruptedGoodeHomolosine",
            "RotatedPole",
            "OSGB",
            "EuroPP",
            "Geostationary",
            "Gnomonic",
            "NorthPolarStereo",
            "SouthPolarStereo",
            "AlbersEqualArea",
            "AzimuthalEquidistant",
            "Sinusoidal",
            "NearsidePerspective",
            "LambertAzimuthalEqualArea",
        ],
    )
    @pytest.mark.mpl_image_compare
    def test_geo_tranform(self, transform):
        field = ("stream", "Density")
        sl = yt.SlicePlot(self.ds, "altitude", field, origin="native")
        sl.set_mpl_projection(transform)
        sl.set_log(field, False)
        sl.render()
        return sl.plots[field].figure



import numpy as np
import pytest

import yt
from yt.testing import fake_hexahedral_ds, fake_tetrahedral_ds, small_fake_hexahedral_ds
from yt.utilities.lib.geometry_utils import triangle_plane_intersect
from yt.utilities.lib.mesh_triangulation import triangulate_indices


def setup_module():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


class TestTetrahedral:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_tetrahedral_ds()
        cls.mesh = cls.ds.index.meshes[0]
        cls.ad = cls.ds.all_data()

        cls.slices = [
            cls.ds.slice(idir, cls.ds.domain_center[idir]) for idir in range(3)
        ]
        cls.sps = [yt.SlicePlot(cls.ds, idir, cls.ds.field_list) for idir in range(3)]
        for sp in cls.sps:
            sp.annotate_mesh_lines()
            sp.set_log("all", False)
            sp.render()

    @pytest.mark.parametrize("idir", range(3))
    def test_mesh_selection(self, idir):
        sl_obj = self.slices[idir]
        for field in self.ds.field_list:
            assert sl_obj[field].shape[0] == self.mesh.count(sl_obj.selector)
            assert sl_obj[field].shape[0] < self.ad[field].shape[0]

    @pytest.mark.parametrize("ax", range(3))
    @pytest.mark.mpl_image_compare
    def test_mesh_slice_tetraheadral_all_elem(self, ax):
        return self.sps[ax].plots["all", "elem"].figure

    @pytest.mark.parametrize("ax", range(3))
    @pytest.mark.mpl_image_compare
    def test_mesh_slice_tetraheadral_all_test(self, ax):
        return self.sps[ax].plots["all", "test"].figure

    @pytest.mark.parametrize("ax", range(3))
    @pytest.mark.mpl_image_compare
    def test_mesh_slice_tetraheadral_connect1_elem(self, ax):
        return self.sps[ax].plots["connect1", "elem"].figure

    @pytest.mark.parametrize("ax", range(3))
    @pytest.mark.mpl_image_compare
    def test_mesh_slice_tetraheadral_connect1_test(self, ax):
        return self.sps[ax].plots["connect1", "test"].figure


class TestHexahedral:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_hexahedral_ds()
        cls.mesh = cls.ds.index.meshes[0]
        cls.ad = cls.ds.all_data()

        cls.slices = [
            cls.ds.slice(idir, cls.ds.domain_center[idir]) for idir in range(3)
        ]
        cls.sps = [yt.SlicePlot(cls.ds, idir, cls.ds.field_list) for idir in range(3)]
        for sp in cls.sps:
            sp.annotate_mesh_lines()
            sp.set_log("all", False)
            sp.render()

    @pytest.mark.parametrize("idir", range(3))
    def test_mesh_selection(self, idir):
        sl_obj = self.slices[idir]
        for field in self.ds.field_list:
            assert sl_obj[field].shape[0] == self.mesh.count(sl_obj.selector)
            assert sl_obj[field].shape[0] < self.ad[field].shape[0]

    @pytest.mark.parametrize("ax", range(3))
    @pytest.mark.mpl_image_compare
    def test_mesh_slice_hexaheadral_all_elem(self, ax):
        return self.sps[ax].plots["all", "elem"].figure

    @pytest.mark.parametrize("ax", range(3))
    @pytest.mark.mpl_image_compare
    def test_mesh_slice_hexaheadral_all_test(self, ax):
        return self.sps[ax].plots["all", "test"].figure

    @pytest.mark.parametrize("ax", range(3))
    @pytest.mark.mpl_image_compare
    def test_mesh_slice_hexaheadral_connect1_elem(self, ax):
        return self.sps[ax].plots["connect1", "elem"].figure

    @pytest.mark.parametrize("ax", range(3))
    @pytest.mark.mpl_image_compare
    def test_mesh_slice_hexaheadral_connect1_test(self, ax):
        return self.sps[ax].plots["connect1", "test"].figure


def test_perfect_element_intersection():
    # This test tests mesh line annotation where a z=0 slice
    # perfectly intersects the top of a hexahedral element with node
    # z-coordinates containing both -0 and +0. Before
    # https://github.com/yt-project/yt/pull/1437 this test falsely
    # yielded three annotation lines, whereas the correct result is four
    # corresponding to the four edges of the top hex face.

    ds = small_fake_hexahedral_ds()
    indices = ds.index.meshes[0].connectivity_indices
    coords = ds.index.meshes[0].connectivity_coords
    tri_indices = triangulate_indices(indices)
    tri_coords = coords[tri_indices]
    lines = triangle_plane_intersect(2, 0, tri_coords)
    non_zero_lines = 0
    for i in range(lines.shape[0]):
        norm = np.linalg.norm(lines[i][0] - lines[i][1])
        if norm > 1e-8:
            non_zero_lines += 1
    np.testing.assert_equal(non_zero_lines, 4)



import yt

# Load the data file.
ds = yt.load("Sedov_3d/sedov_hdf5_chk_0002")

# Make a traditional slice plot.
sp = yt.SlicePlot(ds, "x", ("gas", "density"))

# Overlay the slice plot with thick red contours of density.
sp.annotate_contour(
    ("gas", "density"),
    levels=3,
    clim=(1e-2, 1e-1),
    label=True,
    plot_args={"colors": "red", "linewidths": 2},
)

# What about some nice temperature contours in blue?
sp.annotate_contour(
    ("gas", "temperature"),
    levels=3,
    clim=(1e-8, 1e-6),
    label=True,
    plot_args={"colors": "blue", "linewidths": 2},
)

# This is the plot object.
po = sp.plots["gas", "density"]

# Turn off the colormap image, leaving just the contours.
po.axes.images[0].set_visible(False)

# Remove the colorbar and its label.
po.figure.delaxes(po.figure.axes[1])

# Save it and ask for a close fit to get rid of the space used by the colorbar.
sp.save(mpl_kwargs={"bbox_inches": "tight"})



from .clump_handling import Clump, find_clumps
from .clump_info_items import add_clump_info
from .clump_tools import (
    clump_list_sort,
    recursive_all_clumps,
    recursive_bottom_clumps,
    return_all_clumps,
    return_bottom_clumps,
)
from .clump_validators import add_validator
from .contour_finder import identify_contours



"""
Oct visitor definitions file




"""


cimport numpy as np


cdef struct Oct
cdef struct Oct:
    np.int64_t file_ind     # index with respect to the order in which it was
                            # added
    np.int64_t domain_ind   # index within the global set of domains
    np.int64_t domain       # (opt) addl int index
    Oct **children          # Up to 8 long

cdef struct OctInfo:
    np.float64_t left_edge[3]
    np.float64_t dds[3]
    np.int64_t ipos[3]
    np.int32_t level

cdef struct OctPadded:
    np.int64_t file_ind
    np.int64_t domain_ind
    np.int64_t domain
    np.int64_t padding

cdef class OctVisitor:
    cdef np.uint64_t index
    cdef np.uint64_t last
    cdef np.int64_t global_index
    cdef np.int64_t pos[3]       # position in ints
    cdef np.uint8_t ind[3]       # cell position
    cdef int dims
    cdef np.int32_t domain
    cdef np.int8_t level
    cdef np.int8_t nz # This is number of zones along each dimension.  1 => 1 zones, 2 => 8, etc.
                        # To calculate nzones, nz**3
    cdef np.int32_t nzones

    # There will also be overrides for the memoryviews associated with the
    # specific instance.

    cdef void visit(self, Oct*, np.uint8_t selected)

    cdef inline int oind(self):
        cdef int d = self.nz
        return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])

    cdef inline int rind(self):
        cdef int d = self.nz
        return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])

cdef class CountTotalOcts(OctVisitor):
    pass

cdef class CountTotalCells(OctVisitor):
    pass

cdef class MarkOcts(OctVisitor):
    # Unused
    cdef np.uint8_t[:,:,:,:] mark

cdef class MaskOcts(OctVisitor):
    cdef np.uint8_t[:,:,:,:] mask

cdef class IndexOcts(OctVisitor):
    cdef np.int64_t[:] oct_index

cdef class MaskedIndexOcts(OctVisitor):
    cdef np.int64_t[:] oct_index
    cdef np.uint8_t[:] oct_mask

cdef class IndexMaskMapOcts(OctVisitor):
    cdef np.int64_t[:] oct_index
    cdef np.uint8_t[:] oct_mask
    cdef np.int64_t[:] map_domain_ind
    cdef np.uint64_t map_index

cdef class ICoordsOcts(OctVisitor):
    cdef np.int64_t[:,:] icoords

cdef class IResOcts(OctVisitor):
    cdef np.int64_t[:] ires

cdef class FCoordsOcts(OctVisitor):
    cdef np.float64_t[:,:] fcoords

cdef class FWidthOcts(OctVisitor):
    cdef np.float64_t[:,:] fwidth

cdef class CopyArrayI64(OctVisitor):
    cdef np.int64_t[:,:,:,:,:,:] source
    cdef np.int64_t[:,:] dest

cdef class CopyArrayF64(OctVisitor):
    cdef np.float64_t[:,:,:,:,:] source
    cdef np.float64_t[:,:] dest

cdef class CopyFileIndArrayI8(OctVisitor):
    cdef np.int64_t root
    cdef np.uint8_t[:] source
    cdef np.uint8_t[:] dest

cdef class IdentifyOcts(OctVisitor):
    cdef np.uint8_t[:] domain_mask

cdef class AssignDomainInd(OctVisitor):
    pass

cdef class FillFileIndicesO(OctVisitor):
    cdef np.uint8_t[:] levels
    cdef np.int64_t[:] file_inds
    cdef np.uint8_t[:] cell_inds

cdef class FillFileIndicesR(OctVisitor):
    cdef np.uint8_t[:] levels
    cdef np.int64_t[:] file_inds
    cdef np.uint8_t[:] cell_inds

cdef class CountByDomain(OctVisitor):
    cdef np.int64_t[:] domain_counts

cdef class StoreOctree(OctVisitor):
    cdef np.uint8_t[:] ref_mask

cdef class LoadOctree(OctVisitor):
    cdef np.uint8_t[:] ref_mask
    cdef Oct* octs
    cdef np.uint64_t *nocts
    cdef np.uint64_t *nfinest
    cdef np.uint64_t max_level

cdef class MortonIndexOcts(OctVisitor):
    cdef np.uint8_t[:] level_arr
    cdef np.uint64_t[:] morton_ind

cdef inline int cind(int i, int j, int k) noexcept nogil:
    # THIS ONLY WORKS FOR CHILDREN.  It is not general for zones.
    return (((i*2)+j)*2+k)

from .oct_container cimport OctreeContainer


cdef class StoreIndex(OctVisitor):
    cdef np.int64_t[:,:,:,:] cell_inds

# cimport oct_container
cdef class BaseNeighbourVisitor(OctVisitor):
    cdef int idim      # 0,1,2 for x,y,z
    cdef int direction # +1 for +x, -1 for -x
    cdef np.uint8_t neigh_ind[3]
    cdef bint other_oct
    cdef Oct *neighbour
    cdef OctreeContainer octree
    cdef OctInfo oi
    cdef int n_ghost_zones

    cdef void set_neighbour_info(self, Oct *o, int ishift[3])

    cdef inline np.uint8_t neighbour_rind(self):
        cdef int d = self.nz
        return (((self.neigh_ind[2]*d)+self.neigh_ind[1])*d+self.neigh_ind[0])

cdef class NeighbourCellIndexVisitor(BaseNeighbourVisitor):
    cdef np.uint8_t[::1] cell_inds
    cdef np.int64_t[::1] domain_inds

cdef class NeighbourCellVisitor(BaseNeighbourVisitor):
    cdef np.uint8_t[::1] levels
    cdef np.int64_t[::1] file_inds
    cdef np.uint8_t[::1] cell_inds
    cdef np.int32_t[::1] domains



import yt

# load the dataset
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# create our plot
p = yt.ParticlePlot(
    ds,
    ("all", "particle_position_x"),
    ("all", "particle_velocity_z"),
    [("all", "particle_mass")],
)

# pick some appropriate units
p.set_unit(("all", "particle_position_x"), "Mpc")
p.set_unit(("all", "particle_velocity_z"), "km/s")
p.set_unit(("all", "particle_mass"), "Msun")

# save result
p.save()



import numpy as np

import yt
from yt.data_objects.particle_filters import add_particle_filter


# Define filter functions for our particle filters based on stellar age.
# In this dataset particles in the initial conditions are given creation
# times arbitrarily far into the future, so stars with negative ages belong
# in the old stars filter.
def stars_10Myr(pfilter, data):
    age = data.ds.current_time - data["Stars", "creation_time"]
    filter = np.logical_and(age >= 0, age.in_units("Myr") < 10)
    return filter


def stars_100Myr(pfilter, data):
    age = (data.ds.current_time - data["Stars", "creation_time"]).in_units("Myr")
    filter = np.logical_and(age >= 10, age < 100)
    return filter


def stars_old(pfilter, data):
    age = data.ds.current_time - data["Stars", "creation_time"]
    filter = np.logical_or(age < 0, age.in_units("Myr") >= 100)
    return filter


# Create the particle filters
add_particle_filter(
    "stars_young",
    function=stars_10Myr,
    filtered_type="Stars",
    requires=["creation_time"],
)
add_particle_filter(
    "stars_medium",
    function=stars_100Myr,
    filtered_type="Stars",
    requires=["creation_time"],
)
add_particle_filter(
    "stars_old", function=stars_old, filtered_type="Stars", requires=["creation_time"]
)

# Load a dataset and apply the particle filters
filename = "TipsyGalaxy/galaxy.00300"
ds = yt.load(filename)
ds.add_particle_filter("stars_young")
ds.add_particle_filter("stars_medium")
ds.add_particle_filter("stars_old")

# What are the total masses of different ages of star in the whole simulation
# volume?
ad = ds.all_data()
mass_young = ad["stars_young", "particle_mass"].in_units("Msun").sum()
mass_medium = ad["stars_medium", "particle_mass"].in_units("Msun").sum()
mass_old = ad["stars_old", "particle_mass"].in_units("Msun").sum()
print(f"Mass of young stars = {mass_young:g}")
print(f"Mass of medium stars = {mass_medium:g}")
print(f"Mass of old stars = {mass_old:g}")

# Generate 4 projections: gas density, young stars, medium stars, old stars
fields = [
    ("stars_young", "particle_mass"),
    ("stars_medium", "particle_mass"),
    ("stars_old", "particle_mass"),
]

prj1 = yt.ProjectionPlot(ds, "z", ("gas", "density"), center="max", width=(100, "kpc"))
prj1.save()
prj2 = yt.ParticleProjectionPlot(ds, "z", fields, center="max", width=(100, "kpc"))
prj2.save()



import yt

# Load the dataset.
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# Create a 15 kpc radius sphere, centered on the center of the sim volume
sp = ds.sphere("center", (15.0, "kpc"))

# Get the angular momentum vector for the sphere.
L = sp.quantities.angular_momentum_vector()

print(f"Angular momentum vector: {L}")

# Create an off-axis ProjectionPlot of density centered on the object with the L
# vector as its normal and a width of 25 kpc on a side
p = yt.ProjectionPlot(
    ds, L, fields=("gas", "density"), center=sp.center, width=(25, "kpc")
)
p.save()



__all__ = [
    "adaptahop",
    "ahf",
    "amrex",
    "amrvac",
    "art",
    "artio",
    "athena",
    "athena_pp",
    "boxlib",  # the boxlib frontend is deprecated, use 'amrex'
    "cf_radial",
    "chimera",
    "chombo",
    "cholla",
    "enzo_e",
    "enzo",
    "exodus_ii",
    "fits",
    "flash",
    "gadget",
    # breaking alphabetical order intentionnally here:
    # arepo and eagle depend on gadget. Importing them first causes
    # unintended side effects. See
    # https://github.com/yt-project/yt/issues/4563
    "arepo",
    "eagle",
    "gadget_fof",
    "gamer",
    "gdf",
    "gizmo",
    "halo_catalog",
    "http_stream",
    "moab",
    "nc4_cm1",
    "open_pmd",
    "owls",
    "owls_subfind",
    "parthenon",
    "ramses",
    "rockstar",
    "sdf",
    "stream",
    "swift",
    "tipsy",
    "ytdata",
]

from functools import lru_cache


@lru_cache(maxsize=None)
def __getattr__(value):
    import importlib

    if value == "_all":
        # recursively import all frontends
        for _ in __all__:
            __getattr__(_)
        return

    if value not in __all__:
        raise AttributeError(f"yt.frontends has no attribute {value!r}")

    return importlib.import_module(f"yt.frontends.{value}.api")



import itertools as it
import os
import shutil
import tempfile
import unittest

import numpy as np
from numpy.testing import assert_equal

from yt.testing import (
    assert_fname,
    assert_rel_equal,
    fake_octree_ds,
    fake_random_ds,
)
from yt.visualization.api import (
    OffAxisProjectionPlot,
    OffAxisSlicePlot,
    ProjectionPlot,
)
from yt.visualization.image_writer import write_projection
from yt.visualization.volume_rendering.api import off_axis_projection


# TODO: replace this with pytest.mark.parametrize
def expand_keywords(keywords, full=False):
    """
    expand_keywords is a means for testing all possible keyword
    arguments in the nosetests.  Simply pass it a dictionary of all the
    keyword arguments and all of the values for these arguments that you
    want to test.

    It will return a list of kwargs dicts containing combinations of
    the various kwarg values you passed it.  These can then be passed
    to the appropriate function in nosetests.

    If full=True, then every possible combination of keywords is produced,
    otherwise, every keyword option is included at least once in the output
    list.  Be careful, by using full=True, you may be in for an exponentially
    larger number of tests!

    Parameters
    ----------

    keywords : dict
        a dictionary where the keys are the keywords for the function,
        and the values of each key are the possible values that this key
        can take in the function

    full : bool
        if set to True, every possible combination of given keywords is
        returned

    Returns
    -------

    array of dicts
        An array of dictionaries to be individually passed to the appropriate
        function matching these kwargs.

    Examples
    --------

    >>> keywords = {}
    >>> keywords["dpi"] = (50, 100, 200)
    >>> keywords["cmap"] = ("cmyt.arbre", "cmyt.kelp")
    >>> list_of_kwargs = expand_keywords(keywords)
    >>> print(list_of_kwargs)

    array([{'cmap': 'cmyt.arbre', 'dpi': 50},
           {'cmap': 'cmyt.kelp', 'dpi': 100},
           {'cmap': 'cmyt.arbre', 'dpi': 200}], dtype=object)

    >>> list_of_kwargs = expand_keywords(keywords, full=True)
    >>> print(list_of_kwargs)

    array([{'cmap': 'cmyt.arbre', 'dpi': 50},
           {'cmap': 'cmyt.arbre', 'dpi': 100},
           {'cmap': 'cmyt.arbre', 'dpi': 200},
           {'cmap': 'cmyt.kelp', 'dpi': 50},
           {'cmap': 'cmyt.kelp', 'dpi': 100},
           {'cmap': 'cmyt.kelp', 'dpi': 200}], dtype=object)

    >>> for kwargs in list_of_kwargs:
    ...     write_projection(*args, **kwargs)
    """
    # if we want every possible combination of keywords, use iter magic
    if full:
        keys = sorted(keywords)
        list_of_kwarg_dicts = np.array(
            [
                dict(zip(keys, prod, strict=True))
                for prod in it.product(*(keywords[key] for key in keys))
            ]
        )

    # if we just want to probe each keyword, but not necessarily every
    # combination
    else:
        # Determine the maximum number of values any of the keywords has
        num_lists = 0
        for val in keywords.values():
            if isinstance(val, str):
                num_lists = max(1.0, num_lists)
            else:
                num_lists = max(len(val), num_lists)

        # Construct array of kwargs dicts, each element of the list is a different
        # **kwargs dict.  each kwargs dict gives a different combination of
        # the possible values of the kwargs

        # initialize array
        list_of_kwarg_dicts = np.array([{} for x in range(num_lists)])

        # fill in array
        for i in np.arange(num_lists):
            list_of_kwarg_dicts[i] = {}
            for key in keywords.keys():
                # if it's a string, use it (there's only one)
                if isinstance(keywords[key], str):
                    list_of_kwarg_dicts[i][key] = keywords[key]
                # if there are more options, use the i'th val
                elif i < len(keywords[key]):
                    list_of_kwarg_dicts[i][key] = keywords[key][i]
                # if there are not more options, use the 0'th val
                else:
                    list_of_kwarg_dicts[i][key] = keywords[key][0]

    return list_of_kwarg_dicts


class TestOffAxisProjection(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.mkdtemp()
        cls.curdir = os.getcwd()
        os.chdir(cls.tmpdir)

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.curdir)
        shutil.rmtree(cls.tmpdir)

    def test_oap(self):
        """Tests functionality of off_axis_projection and write_projection."""

        # args for off_axis_projection
        test_ds = fake_random_ds(64)
        c = test_ds.domain_center
        norm = [0.5, 0.5, 0.5]
        W = test_ds.arr([0.5, 0.5, 1.0], "unitary")
        N = 256
        field = ("gas", "density")
        oap_args = [test_ds, c, norm, W, N, field]

        # kwargs for off_axis_projection
        oap_kwargs = {}
        oap_kwargs["weight"] = (None, "cell_mass")
        oap_kwargs["no_ghost"] = (True, False)
        oap_kwargs["interpolated"] = (False,)
        oap_kwargs["north_vector"] = ((1, 0, 0), (0, 0.5, 1.0))
        oap_kwargs_list = expand_keywords(oap_kwargs)

        # args or write_projection
        fn = "test_%d.png"

        # kwargs for write_projection
        wp_kwargs = {}
        wp_kwargs["colorbar"] = (True, False)
        wp_kwargs["colorbar_label"] = "test"
        wp_kwargs["title"] = "test"
        wp_kwargs["vmin"] = (None,)
        wp_kwargs["vmax"] = (1e3, 1e5)
        wp_kwargs["take_log"] = (True, False)
        wp_kwargs["figsize"] = ((8, 6), [1, 1])
        wp_kwargs["dpi"] = (100, 50)
        wp_kwargs["cmap_name"] = ("cmyt.arbre", "cmyt.kelp")
        wp_kwargs_list = expand_keywords(wp_kwargs)

        # test all off_axis_projection kwargs and write_projection kwargs
        # make sure they are able to be projected, then remove and try next
        # iteration
        for i, oap_kwargs in enumerate(oap_kwargs_list):
            image = off_axis_projection(*oap_args, **oap_kwargs)
            for wp_kwargs in wp_kwargs_list:
                write_projection(image, fn % i, **wp_kwargs)
                assert_fname(fn % i)

        # Test remaining parameters of write_projection
        write_projection(image, "test_2", xlabel="x-axis", ylabel="y-axis")
        assert_fname("test_2.png")

        write_projection(image, "test_3.pdf", xlabel="x-axis", ylabel="y-axis")
        assert_fname("test_3.pdf")

        write_projection(image, "test_4.eps", xlabel="x-axis", ylabel="y-axis")
        assert_fname("test_4.eps")


def test_field_cut_off_axis_octree():
    ds = fake_octree_ds()
    cut = ds.all_data().cut_region('obj["gas", "density"]>0.5')
    p1 = OffAxisProjectionPlot(ds, [1, 0, 0], ("gas", "density"))
    p2 = OffAxisProjectionPlot(ds, [1, 0, 0], ("gas", "density"), data_source=cut)
    assert_equal(p2.frb["gas", "density"].min() == 0.0, True)  # Lots of zeros
    assert_equal((p1.frb["gas", "density"] == p2.frb["gas", "density"]).all(), False)
    p3 = OffAxisSlicePlot(ds, [1, 0, 0], ("gas", "density"))
    p4 = OffAxisSlicePlot(ds, [1, 0, 0], ("gas", "density"), data_source=cut)
    assert_equal((p3.frb["gas", "density"] == p4.frb["gas", "density"]).all(), False)
    p4rho = p4.frb["gas", "density"]
    assert_equal(np.nanmin(p4rho[p4rho > 0.0]) >= 0.5, True)


def test_off_axis_octree():
    np.random.seed(12345)
    ds = fake_octree_ds()
    center = [0.4, 0.4, 0.4]

    for weight in [("gas", "cell_mass"), None, ("index", "dx")]:
        p1 = ProjectionPlot(
            ds,
            "x",
            ("gas", "density"),
            center=center,
            width=0.8,
            weight_field=weight,
        )
        p2 = OffAxisProjectionPlot(
            ds,
            [1, 0, 0],
            ("gas", "density"),
            center=center,
            width=0.8,
            weight_field=weight,
        )

        # Note: due to our implementation, the off-axis projection will have a
        # slightly blurred cell edges so we can't do an exact comparison
        v1, v2 = p1.frb["gas", "density"], p2.frb["gas", "density"]
        diff = (v1 - v2) / (v1 + v2) * 2

        # Make sure the difference has a small bias
        assert np.mean(diff).max() < 1e-3  # 0.1%

        # Compute 10-90% percentile
        q10, q90 = np.percentile(diff, q=(10, 90))
        assert q10 > -0.02  # 2%: little up/down deviations
        assert q90 < +0.02  # 2%: little up/down deviations


def test_offaxis_moment():
    ds = fake_random_ds(64)

    def _vlos_sq(field, data):
        return data["gas", "velocity_los"] ** 2

    ds.add_field(
        ("gas", "velocity_los_squared"),
        _vlos_sq,
        sampling_type="local",
        units="cm**2/s**2",
    )
    p1 = OffAxisProjectionPlot(
        ds,
        [1, 1, 1],
        [("gas", "velocity_los"), ("gas", "velocity_los_squared")],
        weight_field=("gas", "density"),
        moment=1,
        buff_size=(400, 400),
    )
    p2 = OffAxisProjectionPlot(
        ds,
        [1, 1, 1],
        ("gas", "velocity_los"),
        weight_field=("gas", "density"),
        moment=2,
        buff_size=(400, 400),
    )
    ## this failed because some <v**2> - <v>**2 values come out
    ## marginally < 0, resulting in unmatched NaN values in the
    ## first assert_rel_equal argument. The compute_stddev_image
    ## function used in OffAxisProjectionPlot checks for and deals
    ## with these cases.
    # assert_rel_equal(
    #    np.sqrt(
    #        p1.frb["gas", "velocity_los_squared"] - p1.frb["gas", "velocity_los"] ** 2
    #    ),
    #    p2.frb["gas", "velocity_los"],
    #    10,
    # )
    p1_expsq = p1.frb["gas", "velocity_los_squared"]
    p1_sqexp = p1.frb["gas", "velocity_los"] ** 2
    # set values to zero that have <v>**2 - <v>**2 < 0, but
    # the absolute values are much smaller than the smallest
    # postive values of <v>**2 and <v>**2
    # (i.e., the difference is pretty much zero)
    mindiff = 1e-10 * min(
        np.min(p1_expsq[p1_expsq > 0]), np.min(p1_sqexp[p1_sqexp > 0])
    )
    # print(mindiff)
    safeorbad = np.logical_not(
        np.logical_and(p1_expsq - p1_sqexp < 0, p1_expsq - p1_sqexp > -1.0 * mindiff)
    )
    # avoid errors from sqrt(negative)
    # sqrt in zeros_like insures correct units
    p1res = np.zeros_like(np.sqrt(p1_expsq))
    p1res[safeorbad] = np.sqrt(p1_expsq[safeorbad] - p1_sqexp[safeorbad])
    p2res = p2.frb["gas", "velocity_los"]
    assert_rel_equal(p1res, p2res, 10)


def test_two_offaxis():
    ds = fake_octree_ds()

    vec_proj = [1, 0, 0]
    los_unit_vector = np.array(vec_proj) / np.linalg.norm(vec_proj)

    # define a new yt velocity field relative to the bulk velocity of the halo and along the line of sight
    def _velocity_los_test(field, data):
        v = np.stack(
            [data["gas", f"velocity_{k}"].to("km/s").d for k in "xyz"], axis=-1
        )
        v_los = np.dot(v, los_unit_vector)
        return data.apply_units(v_los, "km/s")

    ds.add_field(
        ("gas", "velocity_los_test"),
        function=_velocity_los_test,
        units="km/s",
        sampling_type="cell",
        display_name=f"Velocity LOS along {vec_proj}",
    )

    p = OffAxisProjectionPlot(
        ds,
        [1.0, 0.0, 0.0],
        [("gas", "velocity_x"), ("gas", "velocity_los_test")],
        weight_field="dx",
    )

    v1 = p.frb["gas", "velocity_x"]
    v2 = p.frb["gas", "velocity_los_test"]

    np.testing.assert_allclose(v1, v2.to(v1.units))



import pickle

import numpy as np
from numpy.testing import assert_equal

from yt.testing import fake_particle_ds


def test_pickleability():
    # tests the pickleability of the selection objects.

    def pickle_test(sel_obj):
        assert_fields = sel_obj._get_state_attnames()  # the attrs used in get/set state
        new_sel_obj = pickle.loads(pickle.dumps(sel_obj))
        for attr in assert_fields:
            assert_equal(getattr(new_sel_obj, attr), getattr(sel_obj, attr))

    # list of selection types and argument tuples for each selection type
    c = np.array([0.5, 0.5, 0.5])
    sargs = (
        ("point", (c,)),
        ("sphere", (c, 0.25)),
        ("box", (c - 0.3, c)),
        ("ellipsoid", (c, 0.3, 0.2, 0.1, c - 0.4, 0.2)),
        ("disk", (c, [1, 0, 0], 0.2, 0.2)),
        ("cutting", ([0.1, 0.2, -0.9], [0.5, 0.42, 0.6])),
        ("ortho_ray", ("z", c)),
        ("ray", (c, [0.1, 0.1, 0.1])),
    )

    # load fake data
    ds = fake_particle_ds()
    for sel_type, args in sargs:
        sel = getattr(ds, sel_type)(*args)  # instantiate this selection type
        pickle_test(sel.selector)  # make sure it (un)pickles



import numpy as np
from numpy.testing import assert_array_equal, assert_equal

import yt


def setup_fake_refby():
    refine_by = np.array([5, 1, 1])
    top_grid_dim = [100, 10, 2]
    n1 = 100
    n2 = 10
    n3 = 2

    grid_data = [
        {
            "left_edge": [0.0, 0.0, 0.0],
            "right_edge": [1.0, np.pi, np.pi * 2.0],
            "level": 0,
            "dimensions": np.array([n1, n2, n3]),
        },
        {
            "left_edge": [0.0, 0.0, 0.0],
            "right_edge": [0.5, np.pi, np.pi * 2.0],
            "level": 1,
            "dimensions": refine_by * [n1 / 2.0, n2, n3],
        },
    ]

    for g in grid_data:
        g["density"] = (np.random.random(g["dimensions"].astype("i8")), "g/cm**3")
    bbox = np.array([[0.0, 1.0], [0.0, np.pi], [0.0, np.pi * 2]])

    ds = yt.load_amr_grids(
        grid_data,
        top_grid_dim,
        bbox=bbox,
        geometry="spherical",
        refine_by=refine_by,
        length_unit="kpc",
    )
    return ds


def test_refine_by():
    ds = setup_fake_refby()
    dd = ds.all_data()
    # This checks that we always refine_by 1 in dimensions 2 and 3
    dims = ds.domain_dimensions * ds.refine_by**ds.max_level
    for i in range(1, 3):
        # Check the refine_by == 1
        ncoords = np.unique(dd.icoords[:, i]).size
        assert_equal(ncoords, dims[i])
    for g in ds.index.grids:
        dims = ds.domain_dimensions * ds.refine_by**g.Level
        # Now we can check converting back to the reference space
        v = ((g.icoords + 1) / dims.astype("f8")).max(axis=0)
        v *= ds.domain_width
        assert_array_equal(v, g.RightEdge.d)



"""
Created on Wed Feb 18 18:24:09 2015

@author: stuart
"""

from ..hierarchy_inspection import find_lowest_subclasses


class level1:
    pass


class level1a:
    pass


class level2(level1):
    pass


class level3(level2):
    pass


class level4(level3):
    pass


def test_empty():
    result = find_lowest_subclasses([])
    assert len(result) == 0


def test_single():
    result = find_lowest_subclasses([level2])
    assert len(result) == 1
    assert result[0] is level2


def test_two_classes():
    result = find_lowest_subclasses([level1, level2])
    assert len(result) == 1
    assert result[0] is level2


def test_four_deep():
    result = find_lowest_subclasses([level1, level2, level3, level4])
    assert len(result) == 1
    assert result[0] is level4


def test_four_deep_outoforder():
    result = find_lowest_subclasses([level2, level3, level1, level4])
    assert len(result) == 1
    assert result[0] is level4


def test_diverging_tree():
    result = find_lowest_subclasses([level1, level2, level3, level1a])
    assert len(result) == 2
    assert level1a in result and level3 in result


def test_without_parents():
    result = find_lowest_subclasses([level1, level3])
    assert len(result) == 1
    assert result[0] is level3


def test_without_grandparents():
    result = find_lowest_subclasses([level1, level4])
    assert len(result) == 1
    assert result[0] is level4



import argparse
import base64
import json
import os
import pprint
import sys
import textwrap
from typing import Any

import numpy as np
from more_itertools import always_iterable

from yt.config import ytcfg
from yt.funcs import (
    download_file,
    enable_plugins,
    ensure_dir,
    ensure_dir_exists,
    get_git_version,
    mylog,
    update_git,
)
from yt.loaders import load
from yt.utilities.exceptions import YTFieldNotParseable, YTUnidentifiedDataType
from yt.utilities.metadata import get_metadata
from yt.visualization.plot_window import ProjectionPlot, SlicePlot

# isort: off
# This needs to be set before importing startup_tasks
ytcfg["yt", "internals", "command_line"] = True  # isort: skip
from yt.startup_tasks import parser, subparsers  # isort: skip # noqa: E402

# isort: on

# loading field plugins for backward compatibility, since this module
# used to do "from yt.mods import *"

try:
    enable_plugins()
except FileNotFoundError:
    pass

_default_colormap = ytcfg.get("yt", "default_colormap")


def _fix_ds(arg, *args, **kwargs):
    if os.path.isdir(f"{arg}") and os.path.exists(f"{arg}/{arg}"):
        ds = load(f"{arg}/{arg}", *args, **kwargs)
    elif os.path.isdir(f"{arg}.dir") and os.path.exists(f"{arg}.dir/{arg}"):
        ds = load(f"{arg}.dir/{arg}", *args, **kwargs)
    elif arg.endswith(".index"):
        ds = load(arg[:-10], *args, **kwargs)
    else:
        ds = load(arg, *args, **kwargs)
    return ds


def _add_arg(sc, arg):
    if isinstance(arg, str):
        arg = _common_options[arg].copy()
    elif isinstance(arg, tuple):
        exclusive, *args = arg
        if exclusive:
            grp = sc.add_mutually_exclusive_group()
        else:
            grp = sc.add_argument_group()
        for arg in args:
            _add_arg(grp, arg)
        return
    argc = dict(arg.items())
    argnames = []
    if "short" in argc:
        argnames.append(argc.pop("short"))
    if "longname" in argc:
        argnames.append(argc.pop("longname"))
    sc.add_argument(*argnames, **argc)


def _print_failed_source_update(reinstall=False):
    print()
    print("The yt package is not installed from a git repository,")
    print("so you must update this installation manually.")
    if "Continuum Analytics" in sys.version or "Anaconda" in sys.version:
        # see http://stackoverflow.com/a/21318941/1382869 for why we need
        # to check both Continuum *and* Anaconda
        print()
        print("Since it looks like you are using a python installation")
        print("that is managed by conda, you may want to do:")
        print()
        print("    $ conda update yt")
        print()
        print("to update your yt installation.")
        if reinstall:
            print()
            print("To update all of your packages, you can do:")
            print()
            print("    $ conda update --all")
    else:
        print("If you manage your python dependencies with pip, you may")
        print("want to do:")
        print()
        print("    $ python -m pip install -U yt")
        print()
        print("to update your yt installation.")


def _print_installation_information(path):
    import yt

    print()
    print("yt module located at:")
    print(f"    {path}")
    if "YT_DEST" in os.environ:
        spath = os.path.join(os.environ["YT_DEST"], "src", "yt-supplemental")
        if os.path.isdir(spath):
            print("The supplemental repositories are located at:")
            print(f"    {spath}")
    print()
    print("The current version of yt is:")
    print()
    print("---")
    print(f"Version = {yt.__version__}")
    vstring = get_git_version(path)
    if vstring is not None:
        print(f"Changeset = {vstring.strip()}")
    print("---")
    return vstring


class FileStreamer:
    final_size = None
    next_sent = 0
    chunksize = 100 * 1024

    def __init__(self, f, final_size=None):
        location = f.tell()
        f.seek(0, os.SEEK_END)
        self.final_size = f.tell() - location
        f.seek(location)
        self.f = f

    def __iter__(self):
        from tqdm import tqdm

        with tqdm(
            total=self.final_size, desc="Uploading file", unit="B", unit_scale=True
        ) as pbar:
            while self.f.tell() < self.final_size:
                yield self.f.read(self.chunksize)
                pbar.update(self.chunksize)


_subparsers = {None: subparsers}
_subparsers_description = {
    "config": "Get and set configuration values for yt",
}


class YTCommandSubtype(type):
    def __init__(cls, name, b, d):
        type.__init__(cls, name, b, d)
        if cls.name is None:
            return
        if cls.subparser not in _subparsers:
            try:
                description = _subparsers_description[cls.subparser]
            except KeyError:
                description = cls.subparser
            parent_parser = argparse.ArgumentParser(add_help=False)
            p = subparsers.add_parser(
                cls.subparser,
                help=description,
                description=description,
                parents=[parent_parser],
            )
            _subparsers[cls.subparser] = p.add_subparsers(
                title=cls.subparser, dest=cls.subparser
            )
        sp = _subparsers[cls.subparser]
        for name in always_iterable(cls.name):
            sc = sp.add_parser(name, description=cls.description, help=cls.description)
            sc.set_defaults(func=cls.run)
            for arg in cls.args:
                _add_arg(sc, arg)


class YTCommand(metaclass=YTCommandSubtype):
    args: tuple[str | dict[str, Any], ...] = ()
    name: str | list[str] | None = None
    description: str = ""
    aliases = ()
    ndatasets: int = 1
    subparser: str | None = None

    @classmethod
    def run(cls, args):
        self = cls()
        # Check for some things we know; for instance, comma separated
        # field names should be parsed as tuples.
        if getattr(args, "field", None) is not None and "," in args.field:
            if args.field.count(",") > 1:
                raise YTFieldNotParseable(args.field)
            args.field = tuple(_.strip() for _ in args.field.split(","))
        if getattr(args, "weight", None) is not None and "," in args.weight:
            if args.weight.count(",") > 1:
                raise YTFieldNotParseable(args.weight)
            args.weight = tuple(_.strip() for _ in args.weight.split(","))
        # Some commands need to be run repeatedly on datasets
        # In fact, this is the rule and the opposite is the exception
        # BUT, we only want to parse the arguments once.
        if cls.ndatasets > 1:
            self(args)
        else:
            ds_args = getattr(args, "ds", [])
            if len(ds_args) > 1:
                datasets = args.ds
                for ds in datasets:
                    args.ds = ds
                    self(args)
            elif len(ds_args) == 0:
                datasets = []
                self(args)
            else:
                args.ds = getattr(args, "ds", [None])[0]
                self(args)


class GetParameterFiles(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if len(values) == 1:
            datasets = values
        elif len(values) == 2 and namespace.basename is not None:
            datasets = [
                "%s%04i" % (namespace.basename, r)
                for r in range(int(values[0]), int(values[1]), namespace.skip)
            ]
        else:
            datasets = values
        namespace.ds = [_fix_ds(ds) for ds in datasets]


_common_options = {
    "all": {
        "longname": "--all",
        "dest": "reinstall",
        "default": False,
        "action": "store_true",
        "help": (
            "Reinstall the full yt stack in the current location. "
            "This option has been deprecated and will not have any "
            "effect."
        ),
    },
    "ds": {
        "short": "ds",
        "action": GetParameterFiles,
        "nargs": "+",
        "help": "datasets to run on",
    },
    "ods": {
        "action": GetParameterFiles,
        "dest": "ds",
        "nargs": "*",
        "help": "(Optional) datasets to run on",
    },
    "axis": {
        "short": "-a",
        "longname": "--axis",
        "action": "store",
        "type": int,
        "dest": "axis",
        "default": 4,
        "help": "Axis (4 for all three)",
    },
    "log": {
        "short": "-l",
        "longname": "--log",
        "action": "store_true",
        "dest": "takelog",
        "default": True,
        "help": "Use logarithmic scale for image",
    },
    "linear": {
        "longname": "--linear",
        "action": "store_false",
        "dest": "takelog",
        "help": "Use linear scale for image",
    },
    "text": {
        "short": "-t",
        "longname": "--text",
        "action": "store",
        "type": str,
        "dest": "text",
        "default": None,
        "help": "Textual annotation",
    },
    "field": {
        "short": "-f",
        "longname": "--field",
        "action": "store",
        "type": str,
        "dest": "field",
        "default": "density",
        "help": ("Field to color by, use a comma to separate field tuple values"),
    },
    "weight": {
        "short": "-g",
        "longname": "--weight",
        "action": "store",
        "type": str,
        "dest": "weight",
        "default": None,
        "help": (
            "Field to weight projections with, "
            "use a comma to separate field tuple values"
        ),
    },
    "cmap": {
        "longname": "--colormap",
        "action": "store",
        "type": str,
        "dest": "cmap",
        "default": _default_colormap,
        "help": "Colormap name",
    },
    "zlim": {
        "short": "-z",
        "longname": "--zlim",
        "action": "store",
        "type": float,
        "dest": "zlim",
        "default": None,
        "nargs": 2,
        "help": "Color limits (min, max)",
    },
    "dex": {
        "longname": "--dex",
        "action": "store",
        "type": float,
        "dest": "dex",
        "default": None,
        "nargs": 1,
        "help": "Number of dex above min to display",
    },
    "width": {
        "short": "-w",
        "longname": "--width",
        "action": "store",
        "type": float,
        "dest": "width",
        "default": None,
        "help": "Width in specified units",
    },
    "unit": {
        "short": "-u",
        "longname": "--unit",
        "action": "store",
        "type": str,
        "dest": "unit",
        "default": "1",
        "help": "Desired axes units",
    },
    "center": {
        "short": "-c",
        "longname": "--center",
        "action": "store",
        "type": float,
        "dest": "center",
        "default": None,
        "nargs": 3,
        "help": "Center, space separated (-1 -1 -1 for max)",
    },
    "max": {
        "short": "-m",
        "longname": "--max",
        "action": "store_true",
        "dest": "max",
        "default": False,
        "help": "Center the plot on the density maximum",
    },
    "bn": {
        "short": "-b",
        "longname": "--basename",
        "action": "store",
        "type": str,
        "dest": "basename",
        "default": None,
        "help": "Basename of datasets",
    },
    "output": {
        "short": "-o",
        "longname": "--output",
        "action": "store",
        "type": str,
        "dest": "output",
        "default": "frames/",
        "help": "Folder in which to place output images",
    },
    "outputfn": {
        "short": "-o",
        "longname": "--output",
        "action": "store",
        "type": str,
        "dest": "output",
        "default": None,
        "help": "File in which to place output",
    },
    "skip": {
        "short": "-s",
        "longname": "--skip",
        "action": "store",
        "type": int,
        "dest": "skip",
        "default": 1,
        "help": "Skip factor for outputs",
    },
    "proj": {
        "short": "-p",
        "longname": "--projection",
        "action": "store_true",
        "dest": "projection",
        "default": False,
        "help": "Use a projection rather than a slice",
    },
    "maxw": {
        "longname": "--max-width",
        "action": "store",
        "type": float,
        "dest": "max_width",
        "default": 1.0,
        "help": "Maximum width in code units",
    },
    "minw": {
        "longname": "--min-width",
        "action": "store",
        "type": float,
        "dest": "min_width",
        "default": 50,
        "help": "Minimum width in units of smallest dx (default: 50)",
    },
    "nframes": {
        "short": "-n",
        "longname": "--nframes",
        "action": "store",
        "type": int,
        "dest": "nframes",
        "default": 100,
        "help": "Number of frames to generate",
    },
    "slabw": {
        "longname": "--slab-width",
        "action": "store",
        "type": float,
        "dest": "slab_width",
        "default": 1.0,
        "help": "Slab width in specified units",
    },
    "slabu": {
        "short": "-g",
        "longname": "--slab-unit",
        "action": "store",
        "type": str,
        "dest": "slab_unit",
        "default": "1",
        "help": "Desired units for the slab",
    },
    "ptype": {
        "longname": "--particle-type",
        "action": "store",
        "type": int,
        "dest": "ptype",
        "default": 2,
        "help": "Particle type to select",
    },
    "agecut": {
        "longname": "--age-cut",
        "action": "store",
        "type": float,
        "dest": "age_filter",
        "default": None,
        "nargs": 2,
        "help": "Bounds for the field to select",
    },
    "uboxes": {
        "longname": "--unit-boxes",
        "action": "store_true",
        "dest": "unit_boxes",
        "help": "Display heldsul unit boxes",
    },
    "thresh": {
        "longname": "--threshold",
        "action": "store",
        "type": float,
        "dest": "threshold",
        "default": None,
        "help": "Density threshold",
    },
    "dm_only": {
        "longname": "--all-particles",
        "action": "store_false",
        "dest": "dm_only",
        "default": True,
        "help": "Use all particles",
    },
    "grids": {
        "longname": "--show-grids",
        "action": "store_true",
        "dest": "grids",
        "default": False,
        "help": "Show the grid boundaries",
    },
    "time": {
        "longname": "--time",
        "action": "store_true",
        "dest": "time",
        "default": False,
        "help": "Print time in years on image",
    },
    "contours": {
        "longname": "--contours",
        "action": "store",
        "type": int,
        "dest": "contours",
        "default": None,
        "help": "Number of Contours for Rendering",
    },
    "contour_width": {
        "longname": "--contour_width",
        "action": "store",
        "type": float,
        "dest": "contour_width",
        "default": None,
        "help": "Width of gaussians used for rendering.",
    },
    "enhance": {
        "longname": "--enhance",
        "action": "store_true",
        "dest": "enhance",
        "default": False,
        "help": "Enhance!",
    },
    "valrange": {
        "short": "-r",
        "longname": "--range",
        "action": "store",
        "type": float,
        "dest": "valrange",
        "default": None,
        "nargs": 2,
        "help": "Range, space separated",
    },
    "up": {
        "longname": "--up",
        "action": "store",
        "type": float,
        "dest": "up",
        "default": None,
        "nargs": 3,
        "help": "Up, space separated",
    },
    "viewpoint": {
        "longname": "--viewpoint",
        "action": "store",
        "type": float,
        "dest": "viewpoint",
        "default": [1.0, 1.0, 1.0],
        "nargs": 3,
        "help": "Viewpoint, space separated",
    },
    "pixels": {
        "longname": "--pixels",
        "action": "store",
        "type": int,
        "dest": "pixels",
        "default": None,
        "help": "Number of Pixels for Rendering",
    },
    "halos": {
        "longname": "--halos",
        "action": "store",
        "type": str,
        "dest": "halos",
        "default": "multiple",
        "help": "Run halo profiler on a 'single' halo or 'multiple' halos.",
    },
    "halo_radius": {
        "longname": "--halo_radius",
        "action": "store",
        "type": float,
        "dest": "halo_radius",
        "default": 0.1,
        "help": "Constant radius for profiling halos if using hop output files with no "
        + "radius entry. Default: 0.1.",
    },
    "halo_radius_units": {
        "longname": "--halo_radius_units",
        "action": "store",
        "type": str,
        "dest": "halo_radius_units",
        "default": "1",
        "help": "Units for radius used with --halo_radius flag. "
        + "Default: '1' (code units).",
    },
    "halo_hop_style": {
        "longname": "--halo_hop_style",
        "action": "store",
        "type": str,
        "dest": "halo_hop_style",
        "default": "new",
        "help": "Style of hop output file. "
        + "'new' for yt_hop files and 'old' for enzo_hop files.",
    },
    "halo_dataset": {
        "longname": "--halo_dataset",
        "action": "store",
        "type": str,
        "dest": "halo_dataset",
        "default": None,
        "help": "HaloProfiler dataset.",
    },
    "make_profiles": {
        "longname": "--make_profiles",
        "action": "store_true",
        "default": False,
        "help": "Make profiles with halo profiler.",
    },
    "make_projections": {
        "longname": "--make_projections",
        "action": "store_true",
        "default": False,
        "help": "Make projections with halo profiler.",
    },
}


class YTInstInfoCmd(YTCommand):
    name = ["instinfo", "version"]
    args = (
        {
            "short": "-u",
            "longname": "--update-source",
            "action": "store_true",
            "default": False,
            "help": "Update the yt installation, if able",
        },
        {
            "short": "-o",
            "longname": "--output-version",
            "action": "store",
            "default": None,
            "dest": "outputfile",
            "help": "File into which the current revision number will be stored",
        },
    )
    description = """
        Get some information about the yt installation

        """

    def __call__(self, opts):
        import importlib.resources as importlib_resources

        path = os.path.dirname(importlib_resources.files("yt"))
        vstring = _print_installation_information(path)
        if vstring is not None:
            print("This installation CAN be automatically updated.")
            if opts.update_source:
                update_git(path)
        elif opts.update_source:
            _print_failed_source_update()
        if vstring is not None and opts.outputfile is not None:
            open(opts.outputfile, "w").write(vstring)


class YTLoadCmd(YTCommand):
    name = "load"
    description = """
        Load a single dataset into an IPython instance

        """

    args = ("ds",)

    def __call__(self, args):
        if args.ds is None:
            print("Could not load file.")
            sys.exit()
        import IPython

        import yt

        local_ns = {}
        local_ns["ds"] = args.ds
        local_ns["pf"] = args.ds
        local_ns["yt"] = yt

        try:
            from traitlets.config.loader import Config
        except ImportError:
            from IPython.config.loader import Config

        cfg = Config()
        # prepend sys.path with current working directory
        sys.path.insert(0, "")
        IPython.embed(config=cfg, user_ns=local_ns)


class YTMapserverCmd(YTCommand):
    args = (
        "proj",
        "field",
        "weight",
        "linear",
        "center",
        "width",
        "cmap",
        {
            "short": "-a",
            "longname": "--axis",
            "action": "store",
            "type": int,
            "dest": "axis",
            "default": 0,
            "help": "Axis",
        },
        {
            "short": "-o",
            "longname": "--host",
            "action": "store",
            "type": str,
            "dest": "host",
            "default": None,
            "help": "IP Address to bind on",
        },
        {"short": "ds", "nargs": 1, "type": str, "help": "The dataset to load."},
    )

    name = "mapserver"
    description = """
        Serve a plot in a GMaps-style interface

        """

    def __call__(self, args):
        from yt.frontends.ramses.data_structures import RAMSESDataset
        from yt.visualization.mapserver.pannable_map import PannableMapServer

        # For RAMSES datasets, use the bbox feature to make the dataset load faster
        if RAMSESDataset._is_valid(args.ds) and args.center and args.width:
            kwa = {
                "bbox": [
                    [c - args.width / 2 for c in args.center],
                    [c + args.width / 2 for c in args.center],
                ]
            }
        else:
            kwa = {}

        ds = _fix_ds(args.ds, **kwa)
        if args.center and args.width:
            center = args.center
            width = args.width
            ad = ds.box(
                left_edge=[c - args.width / 2 for c in args.center],
                right_edge=[c + args.width / 2 for c in args.center],
            )
        else:
            center = [0.5] * 3
            width = 1.0
            ad = ds.all_data()

        if args.axis >= 4:
            print("Doesn't work with multiple axes!")
            return
        if args.projection:
            p = ProjectionPlot(
                ds,
                args.axis,
                args.field,
                weight_field=args.weight,
                data_source=ad,
                center=center,
                width=width,
            )
        else:
            p = SlicePlot(
                ds, args.axis, args.field, data_source=ad, center=center, width=width
            )
        p.set_log("all", args.takelog)
        p.set_cmap("all", args.cmap)

        PannableMapServer(p.data_source, args.field, args.takelog, args.cmap)
        try:
            import bottle
        except ImportError as e:
            raise ImportError(
                "The mapserver functionality requires the bottle "
                "package to be installed. Please install using `pip "
                "install bottle`."
            ) from e
        bottle.debug(True)
        if args.host is not None:
            colonpl = args.host.find(":")
            if colonpl >= 0:
                port = int(args.host.split(":")[-1])
                args.host = args.host[:colonpl]
            else:
                port = 8080
            bottle.run(server="auto", host=args.host, port=port)
        else:
            bottle.run(server="auto")


class YTPastebinCmd(YTCommand):
    name = "pastebin"
    args = (
        {
            "short": "-l",
            "longname": "--language",
            "action": "store",
            "default": None,
            "dest": "language",
            "help": "Use syntax highlighter for the file in language",
        },
        {
            "short": "-L",
            "longname": "--languages",
            "action": "store_true",
            "default": False,
            "dest": "languages",
            "help": "Retrieve a list of supported languages",
        },
        {
            "short": "-e",
            "longname": "--encoding",
            "action": "store",
            "default": "utf-8",
            "dest": "encoding",
            "help": "Specify the encoding of a file (default is "
            "utf-8 or guessing if available)",
        },
        {
            "short": "-b",
            "longname": "--open-browser",
            "action": "store_true",
            "default": False,
            "dest": "open_browser",
            "help": "Open the paste in a web browser",
        },
        {
            "short": "-p",
            "longname": "--private",
            "action": "store_true",
            "default": False,
            "dest": "private",
            "help": "Paste as private",
        },
        {
            "short": "-c",
            "longname": "--clipboard",
            "action": "store_true",
            "default": False,
            "dest": "clipboard",
            "help": "File to output to; else, print.",
        },
        {"short": "file", "type": str},
    )
    description = """
        Post a script to an anonymous pastebin

        """

    def __call__(self, args):
        from yt.utilities import lodgeit as lo

        lo.main(
            args.file,
            languages=args.languages,
            language=args.language,
            encoding=args.encoding,
            open_browser=args.open_browser,
            private=args.private,
            clipboard=args.clipboard,
        )


class YTPastebinGrabCmd(YTCommand):
    args = ({"short": "number", "type": str},)
    name = "pastebin_grab"
    description = """
        Print an online pastebin to STDOUT for local use.
        """

    def __call__(self, args):
        from yt.utilities import lodgeit as lo

        lo.main(None, download=args.number)


class YTPlotCmd(YTCommand):
    args = (
        "width",
        "unit",
        "bn",
        "proj",
        "center",
        "zlim",
        "axis",
        "field",
        "weight",
        "skip",
        "cmap",
        "output",
        "grids",
        "time",
        "ds",
        "max",
        "log",
        "linear",
        {
            "short": "-fu",
            "longname": "--field-unit",
            "action": "store",
            "type": str,
            "dest": "field_unit",
            "default": None,
            "help": "Desired field units",
        },
        {
            "longname": "--show-scale-bar",
            "action": "store_true",
            "help": "Annotate the plot with the scale",
        },
    )

    name = "plot"

    description = """
        Create a set of images

        """

    def __call__(self, args):
        ds = args.ds
        center = args.center
        if args.center == (-1, -1, -1):
            mylog.info("No center fed in; seeking.")
            v, center = ds.find_max("density")
        if args.max:
            v, center = ds.find_max("density")
        elif args.center is None:
            center = 0.5 * (ds.domain_left_edge + ds.domain_right_edge)
        center = np.array(center)
        if ds.dimensionality < 3:
            dummy_dimensions = np.nonzero(ds.index.grids[0].ActiveDimensions <= 1)
            axes = dummy_dimensions[0][0]
        elif args.axis == 4:
            axes = range(3)
        else:
            axes = args.axis

        unit = args.unit
        if unit is None:
            unit = "unitary"
        if args.width is None:
            width = None
        else:
            width = (args.width, args.unit)

        for ax in always_iterable(axes):
            mylog.info("Adding plot for axis %i", ax)
            if args.projection:
                plt = ProjectionPlot(
                    ds,
                    ax,
                    args.field,
                    center=center,
                    width=width,
                    weight_field=args.weight,
                )
            else:
                plt = SlicePlot(ds, ax, args.field, center=center, width=width)
            if args.grids:
                plt.annotate_grids()
            if args.time:
                plt.annotate_timestamp()
            if args.show_scale_bar:
                plt.annotate_scale()

            if args.field_unit:
                plt.set_unit(args.field, args.field_unit)

            plt.set_cmap(args.field, args.cmap)
            plt.set_log(args.field, args.takelog)
            if args.zlim:
                plt.set_zlim(args.field, *args.zlim)
            ensure_dir_exists(args.output)
            plt.save(os.path.join(args.output, f"{ds}"))


class YTRPDBCmd(YTCommand):
    name = "rpdb"
    description = """
        Connect to a currently running (on localhost) rpd session.

        Commands run with --rpdb will trigger an rpdb session with any
        uncaught exceptions.

        """
    args = (
        {
            "short": "-t",
            "longname": "--task",
            "action": "store",
            "default": 0,
            "dest": "task",
            "help": "Open a web browser.",
        },
    )

    def __call__(self, args):
        from . import rpdb

        rpdb.run_rpdb(int(args.task))


class YTStatsCmd(YTCommand):
    args = (
        "outputfn",
        "bn",
        "skip",
        "ds",
        "field",
        {
            "longname": "--max",
            "action": "store_true",
            "default": False,
            "dest": "max",
            "help": "Display maximum of field requested through -f option.",
        },
        {
            "longname": "--min",
            "action": "store_true",
            "default": False,
            "dest": "min",
            "help": "Display minimum of field requested through -f option.",
        },
    )
    name = "stats"
    description = """
        Print stats and max/min value of a given field (if requested),
        for one or more datasets

        (default field is density)

        """

    def __call__(self, args):
        ds = args.ds
        ds.print_stats()
        vals = {}
        field = ds._get_field_info(args.field)
        if args.max:
            vals["max"] = ds.find_max(field)
            print(f"Maximum {field.name}: {vals['max'][0]:0.5e} at {vals['max'][1]}")
        if args.min:
            vals["min"] = ds.find_min(field)
            print(f"Minimum {field.name}: {vals['min'][0]:0.5e} at {vals['min'][1]}")
        if args.output is not None:
            t = ds.current_time.to("yr")
            with open(args.output, "a") as f:
                f.write(f"{ds} ({t:0.5e})\n")
                if "min" in vals:
                    f.write(
                        f"Minimum {field.name} is {vals['min'][0]:0.5e} at {vals['min'][1]}\n"
                    )
                if "max" in vals:
                    f.write(
                        f"Maximum {field.name} is {vals['max'][0]:0.5e} at {vals['max'][1]}\n"
                    )


class YTUpdateCmd(YTCommand):
    args = ("all",)
    name = "update"
    description = """
        Update the yt installation to the most recent version

        """

    def __call__(self, opts):
        import importlib.resources as importlib_resources

        path = os.path.dirname(importlib_resources.files("yt"))
        vstring = _print_installation_information(path)
        if vstring is not None:
            print()
            print("This installation CAN be automatically updated.")
            update_git(path)
        else:
            _print_failed_source_update(opts.reinstall)


class YTDeleteImageCmd(YTCommand):
    args = ({"short": "delete_hash", "type": str},)
    description = """
        Delete image from imgur.com.

        """
    name = "delete_image"

    def __call__(self, args):
        import urllib.error
        import urllib.request

        headers = {"Authorization": f"Client-ID {ytcfg.get('yt', 'imagebin_api_key')}"}

        delete_url = ytcfg.get("yt", "imagebin_delete_url")
        req = urllib.request.Request(
            delete_url.format(delete_hash=args.delete_hash),
            headers=headers,
            method="DELETE",
        )
        try:
            response = urllib.request.urlopen(req).read().decode()
        except urllib.error.HTTPError as e:
            print("ERROR", e)
            return {"deleted": False}

        rv = json.loads(response)
        if "success" in rv and rv["success"]:
            print("\nImage successfully deleted!\n")
        else:
            print()
            print("Something has gone wrong!  Here is the server response:")
            print()
            pprint.pprint(rv)


class YTUploadImageCmd(YTCommand):
    args = ({"short": "file", "type": str},)
    description = """
        Upload an image to imgur.com.  Must be PNG.

        """
    name = "upload_image"

    def __call__(self, args):
        import urllib.error
        import urllib.parse
        import urllib.request

        filename = args.file
        if not filename.endswith(".png"):
            print("File must be a PNG file!")
            return 1
        headers = {"Authorization": f"Client-ID {ytcfg.get('yt', 'imagebin_api_key')}"}

        image_data = base64.b64encode(open(filename, "rb").read())
        parameters = {
            "image": image_data,
            type: "base64",
            "name": filename,
            "title": f"{filename} uploaded by yt",
        }
        data = urllib.parse.urlencode(parameters).encode("utf-8")
        req = urllib.request.Request(
            ytcfg.get("yt", "imagebin_upload_url"), data=data, headers=headers
        )
        try:
            response = urllib.request.urlopen(req).read().decode()
        except urllib.error.HTTPError as e:
            print("ERROR", e)
            return {"uploaded": False}
        rv = json.loads(response)
        if "data" in rv and "link" in rv["data"]:
            print()
            print("Image successfully uploaded!  You can find it at:")
            print(f"    {rv['data']['link']}")
            print()
            print("If you'd like to delete it, use the following")
            print(f"    yt delete_image {rv['data']['deletehash']}")
            print()
        else:
            print()
            print("Something has gone wrong!  Here is the server response:")
            print()
            pprint.pprint(rv)


class YTUploadFileCmd(YTCommand):
    args = ({"short": "file", "type": str},)
    description = """
        Upload a file to yt's curldrop.

        """
    name = "upload"

    def __call__(self, args):
        from yt.utilities.on_demand_imports import _requests as requests

        fs = iter(FileStreamer(open(args.file, "rb")))
        upload_url = ytcfg.get("yt", "curldrop_upload_url")
        r = requests.put(upload_url + "/" + os.path.basename(args.file), data=fs)
        print()
        print(r.text)


class YTConfigLocalConfigHandler:
    def load_config(self, args) -> None:
        import os

        from yt.config import YTConfig
        from yt.utilities.configure import CONFIG

        local_config_file = YTConfig.get_local_config_file()
        global_config_file = YTConfig.get_global_config_file()

        local_exists = os.path.exists(local_config_file)
        global_exists = os.path.exists(global_config_file)

        local_arg_exists = hasattr(args, "local")
        global_arg_exists = hasattr(args, "global")

        config_file: str | None = None
        if getattr(args, "local", False):
            config_file = local_config_file
        elif getattr(args, "global", False):
            config_file = global_config_file
        else:
            if local_exists and global_exists:
                s = (
                    "Yt detected a local and a global configuration file, refusing "
                    "to proceed.\n"
                    f"Local config file: {local_config_file}\n"
                    f"Global config file: {global_config_file}"
                )
                # Only print the info about "--global" and "--local" if they exist
                if local_arg_exists and global_arg_exists:
                    s += (
                        "\n"  # missing eol from previous string
                        "Specify which one you want to use using the `--local` or the "
                        "`--global` flags."
                    )
                sys.exit(s)
            elif local_exists:
                config_file = local_config_file
            elif global_exists:
                config_file = global_config_file

            if config_file is None:
                print("WARNING: no configuration file installed.", file=sys.stderr)
            else:
                print(
                    f"INFO: reading configuration file: {config_file}", file=sys.stderr
                )
        CONFIG.read(config_file)

        self.config_file = config_file


_global_local_args = (
    {
        "short": "--local",
        "action": "store_true",
        "help": "Store the configuration in the local configuration file.",
    },
    {
        "short": "--global",
        "action": "store_true",
        "help": "Store the configuration in the global configuration file.",
    },
)


class YTConfigGetCmd(YTCommand, YTConfigLocalConfigHandler):
    subparser = "config"
    name = "get"
    description = "get a config value"
    args = (
        {"short": "section", "help": "The section containing the option."},
        {"short": "option", "help": "The option to retrieve."},
        *_global_local_args,
    )

    def __call__(self, args):
        from yt.utilities.configure import get_config

        self.load_config(args)

        print(get_config(args.section, args.option))


class YTConfigSetCmd(YTCommand, YTConfigLocalConfigHandler):
    subparser = "config"
    name = "set"
    description = "set a config value"
    args = (
        {"short": "section", "help": "The section containing the option."},
        {"short": "option", "help": "The option to set."},
        {"short": "value", "help": "The value to set the option to."},
        *_global_local_args,
    )

    def __call__(self, args):
        from yt.utilities.configure import set_config

        self.load_config(args)
        if self.config_file is None:
            self.config_file = os.path.join(os.getcwd(), "yt.toml")
            print(
                f"INFO: configuration will be written to {self.config_file}",
                file=sys.stderr,
            )
        set_config(args.section, args.option, args.value, self.config_file)


class YTConfigRemoveCmd(YTCommand, YTConfigLocalConfigHandler):
    subparser = "config"
    name = "rm"
    description = "remove a config option"
    args = (
        {"short": "section", "help": "The section containing the option."},
        {"short": "option", "help": "The option to remove."},
        *_global_local_args,
    )

    def __call__(self, args):
        from yt.utilities.configure import rm_config

        self.load_config(args)

        rm_config(args.section, args.option, self.config_file)


class YTConfigListCmd(YTCommand, YTConfigLocalConfigHandler):
    subparser = "config"
    name = "list"
    description = "show the config content"
    args = _global_local_args

    def __call__(self, args):
        from yt.utilities.configure import write_config

        self.load_config(args)

        write_config(sys.stdout)


class YTConfigPrintPath(YTCommand, YTConfigLocalConfigHandler):
    subparser = "config"
    name = "print-path"
    description = "show path to the config file"
    args = _global_local_args

    def __call__(self, args):
        self.load_config(args)

        print(self.config_file)


class YTSearchCmd(YTCommand):
    args = (
        {
            "short": "-o",
            "longname": "--output",
            "action": "store",
            "type": str,
            "dest": "output",
            "default": "yt_index.json",
            "help": "File in which to place output",
        },
        {
            "longname": "--check-all",
            "short": "-a",
            "help": "Attempt to load every file",
            "action": "store_true",
            "default": False,
            "dest": "check_all",
        },
        {
            "longname": "--full",
            "short": "-f",
            "help": "Output full contents of parameter file",
            "action": "store_true",
            "default": False,
            "dest": "full_output",
        },
    )
    description = """
        Attempt to find outputs that yt can recognize in directories.
        """
    name = "search"

    def __call__(self, args):
        from yt.utilities.object_registries import output_type_registry

        candidates = []
        for base, dirs, files in os.walk(".", followlinks=True):
            print("(% 10i candidates) Examining %s" % (len(candidates), base))
            recurse = []
            if args.check_all:
                candidates.extend([os.path.join(base, _) for _ in files])
            for _, otr in sorted(output_type_registry.items()):
                c, r = otr._guess_candidates(base, dirs, files)
                candidates.extend([os.path.join(base, _) for _ in c])
                recurse.append(r)
            if len(recurse) > 0 and not all(recurse):
                del dirs[:]
        # Now we have a ton of candidates.  We're going to do something crazy
        # and try to load each one.
        records = []
        for i, c in enumerate(sorted(candidates)):
            print("(% 10i/% 10i) Evaluating %s" % (i, len(candidates), c))
            try:
                record = get_metadata(c, args.full_output)
            except YTUnidentifiedDataType:
                continue
            records.append(record)
        with open(args.output, "w") as f:
            json.dump(records, f, indent=4)
        print(f"Identified {len(records)} records output to {args.output}")


class YTDownloadData(YTCommand):
    args = (
        {
            "short": "filename",
            "action": "store",
            "type": str,
            "help": "The name of the file to download",
            "nargs": "?",
            "default": "",
        },
        {
            "short": "location",
            "action": "store",
            "type": str,
            "nargs": "?",
            "help": "The location in which to place the file, can be "
            '"supp_data_dir", "test_data_dir", or any valid '
            "path on disk. ",
            "default": "",
        },
        {
            "longname": "--overwrite",
            "short": "-c",
            "help": "Overwrite existing file.",
            "action": "store_true",
            "default": False,
        },
        {
            "longname": "--list",
            "short": "-l",
            "help": "Display all available files.",
            "action": "store_true",
            "default": False,
        },
    )
    description = """
        Download a file from http://yt-project.org/data and save it to a
        particular location. Files can be saved to the locations provided
        by the "test_data_dir" or "supp_data_dir" configuration entries, or
        any valid path to a location on disk.
        """
    name = "download"

    def __call__(self, args):
        if args.list:
            self.get_list()
            return
        if not args.filename:
            raise RuntimeError(
                "You need to provide a filename. See --help "
                "for details or use --list to get available "
                "datasets."
            )
        elif not args.location:
            raise RuntimeError(
                "You need to specify download location. See --help for details."
            )
        data_url = f"http://yt-project.org/data/{args.filename}"
        if args.location in ["test_data_dir", "supp_data_dir"]:
            data_dir = ytcfg.get("yt", args.location)
            if data_dir == "/does/not/exist":
                raise RuntimeError(f"'{args.location}' is not configured!")
        else:
            data_dir = args.location
        if not os.path.exists(data_dir):
            print(f"The directory '{data_dir}' does not exist. Creating...")
            ensure_dir(data_dir)
        data_file = os.path.join(data_dir, args.filename)
        if os.path.exists(data_file) and not args.overwrite:
            raise OSError(f"File '{data_file}' exists and overwrite=False!")
        print(f"Attempting to download file: {args.filename}")
        fn = download_file(data_url, data_file)

        if not os.path.exists(fn):
            raise OSError(f"The file '{args.filename}' did not download!!")
        print(f"File: {args.filename} downloaded successfully to {data_file}")

    def get_list(self):
        import urllib.request

        data = (
            urllib.request.urlopen("http://yt-project.org/data/datafiles.json")
            .read()
            .decode("utf8")
        )
        data = json.loads(data)
        for key in data:
            for ds in data[key]:
                ds["fullname"] = ds["url"].replace("http://yt-project.org/data/", "")
                print("{fullname} ({size}) type: {code}".format(**ds))
                for line in textwrap.wrap(ds["description"]):
                    print("\t", line)


def run_main():
    args = parser.parse_args()
    # The following is a workaround for a nasty Python 3 bug:
    # http://bugs.python.org/issue16308
    # http://bugs.python.org/issue9253
    try:
        args.func
    except AttributeError:
        parser.print_help()
        sys.exit(0)

    args.func(args)


if __name__ == "__main__":
    run_main()
